# -*- coding: utf-8 -*-
"""
Created on Sat Jun 15 21:01:26 2019

@author: yoelr
"""
from .._unit import Outs, Stream, MissingStream, Unit

__all__ = ('Static',)

class Static(Unit):
    """A Static Unit object expects only one input stream (the `feed`) and only one output stream (the `product`). The static metaclass makes the `feed` and `product` share data on material flow rates. Therefore `feed`.mol and `product`.mol will always be equal. For Static subclasses, the `product`'s T, P, and phase will be equal to the `feed`'s upon simulation if no `_run` method is implemented."""
    _N_ins = _N_outs = 1
    def _init_outs(self, outs):
        """Initialize output streams."""
        if outs is None:
            self._outs = Outs(self, (Stream.proxy(None) for i in self._N_outs))
        elif not outs:
            self._outs = Outs(self, (Stream.proxy('') for i in self._ins))
        elif isinstance(outs, Stream):
            self._outs = Outs(self, (outs,))
        elif isinstance(outs, str):
            self._outs = Outs(self, (Stream.proxy(outs),))
        else:
            self._outs = Outs(self, (o if isinstance(o, Stream)
                                     else Stream.proxy(o)
                                     for o in outs))
            
    def link_streams(self):
        """Link product to feed."""
        try: self._outs[0].link = self._ins[0]
        except Exception as Error:
            if MissingStream in (self._ins + self._outs):
                raise AttributeError(f'missing stream object in {repr(self)}')
       
    def simulate(self):
        """Run rigourous simulation and determine all design requirements and costs."""
        self.link_streams()
        super().simulate()
    
    def _run(self):
        o = self._outs[0]
        i = self._ins[0]
        o.T = i.T
        o.P = i.P
        o._phase = i._phase
            
    def _info(self, T, P, flow, fraction):
        """Information on unit."""
        self.link_streams()
        return super()._info(T, P, flow, fraction)
    
    
    


# -*- coding: utf-8 -*-
"""
Created on Sun Jul 21 22:15:30 2019

@author: yoelr
"""

from numpy import asarray, array
from biosteam._utils import wegstein_secant, aitken_secant, wegstein, aitken

__all__ = ('DewPoint',)

class DewPoint:
    __slots__ = ('gamma', 'P', 'T', 'x')
    rootsolver = staticmethod(aitken_secant)
    itersolver = staticmethod(aitken)
    def __init__(self, gamma):
        self.gamma = gamma
    
    def _x_iter_at_P(self, x, T, zP, VPs):
        return zP/(array([i(T) for i in VPs]) * self.gamma(x/x.sum(), T))
    
    def _x_iter_at_T(self, x, T, P, z, Psat):
        return z*P/(Psat * self.gamma(x/x.sum(), T))
    
    def _T_error(self, T, zP, VPs):
        self.x = self.itersolver(self._x_iter_at_P, self.x, 1e-5, args=(T, zP, VPs))
        return 1 - self.x.sum()
    
    def _P_error(self, P, T, z, Psat):
        self.x = self.itersolver(self._x_iter_at_T, self.x, 1e-5, args=(T, P, z, Psat))
        return 1 - self.x.sum()
    
    def solve_Tx(self, z, P):
        """Dew point given composition and pressure.

        **Parameters**

            **y:** [array_like] Vapor phase composition.

            **P:** [float] Pressure (Pa).

        **Returns**

            **T:** [float] Dew point temperature (K).

            **x:** [numpy array] Liquid phase composition.

        >>> from biosteam import Species, DewPoint, Dortmund
        >>> gamma = Dortmund(*Species('Ethanol', 'Water'))
        >>> dp = DewPoint(gamma)
        >>> dp.solve_Tx(z=(0.5, 0.5), P=101325)
        (357.45184742263075, array([0.151, 0.849]))
        """
        z = asarray(z)
        self.P = P
        try:
            self.T = self.rootsolver(self._T_error, self.T, self.T-0.01, 1e-6,
                                     args=(P*z, [s.VaporPressure for s in self.gamma.species]))
        except:
            self.x = z.copy()
            T = (z * [s.Tb for s in self.gamma.species]).sum()
            try:
                self.T = self.rootsolver(self._T_error, T, T-0.01, 1e-6,
                                         args=(P*z, [s.VaporPressure for s in self.gamma.species]))
            except:
                self.x = z.copy()
                T_guess = min([s.Tb for s in self.gamma.species])
                try:
                    self.T = self.rootsolver(self._T_error, T_guess, T_guess-0.01, 1e-6,
                                             args=(P*z, [s.VaporPressure for s in self.gamma.species]))
                except:
                    self.T = T
                    self.x = z.copy()
                
        self.x = self.x/self.x.sum()
        return self.T, self.x
    
    def solve_Px(self, z, T):
        """Dew point given composition and temperature.

        **Parameters**

            **y:** [array_like] Vapor phase composition.

            **T:** [float] Temperature (K).

        **Returns**

            **P:** [float] Dew point pressure (Pa).

            **x:** [numpy array] Liquid phase composition.

        >>> from biosteam import Species, DewPoint, Dortmund
        >>> gamma = Dortmund(*Species('Ethanol', 'Water'))
        >>> dp = DewPoint(gamma)
        >>> dp.solve_Px(z=(0.703, 0.297), T=352.28)
        (111366.15384513882, array([0.6, 0.4]))
 
       """
        z = asarray(z)
        Psat = array([i.VaporPressure(T) for i in self.gamma.species])
        self.T = T
        try:
            self.P = self.rootsolver(self._P_error, self.P, self.P+1, 1e-2,
                                     args=(T, z, Psat))
        except:
            P = (z * Psat).sum()
            self.x = z.copy()
            self.P = self.rootsolver(self._P_error, P, P+1, 1e-2,
                                     args=(T, z, Psat))
        self.x = self.x/self.x.sum()
        return self.P, self.x
    
    def __repr__(self):
        return f"<{type(self).__name__}: gamma={self.gamma}>"


# -*- coding: utf-8 -*-
"""
Created on Mon Mar  4 11:32:01 2019

@author: yoelr
"""

from .. import Unit
from .decorators import cost
from ._static import Static

@cost('Flow rate', units='kg/hr', CE=576, cost=533471, S=333333, n=0.6)
class MagneticSeparator(Static): pass
    


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:07:01 2018

@author: yoelr
"""
from ._tank import MixTank
from ._hx import HXutility

class EnzymeTreatment(MixTank):
    """Create an EnzymeTreatment unit that is cost as a MixTank with a heat exchanger."""
    _N_outs = 1
    
    #: Residence time (hr)
    _tau = 1
    
    def __init__(self, ID='', ins=None, outs=(), *, T):
        super().__init__(ID, ins, outs)
        self.T = T #: Operating temperature
        self._heat_exchanger = he = HXutility(None, None, T=T) 
        self._heat_utilities = he._heat_utilities
        he._ins = self._ins
        he._outs = self._outs
    
    def _run(self):
        feed = self.ins[0]
        out = self.outs[0]
        out._mol[:] = self._mol_in
        out.phase = feed.phase
        out.P = feed.P
        out.T = self.T
        
    def _design(self):
        super()._design()
        self._heat_exchanger._design()
        
    def _cost(self):
        super()._cost()
        he = self._heat_exchanger
        he._cost()
        self._Cost['Heat exchanger'] = he._Cost['Heat exchanger'] 
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Mon Apr 16 11:49:52 2018

This module includes classes and functions relating exception handling.

@author: Yoel Rene Cortes-Pena
"""
import sys as _sys

__all__ = ('DesignError', 'SolverError', 'EquilibriumError','DimensionError', 'DesignWarning')

# %% Biosteam errors

class DesignError(RuntimeError):
    """RuntimeError regarding unit design."""

class SolverError(RuntimeError):
    """RuntimeError regarding solvers."""

class EquilibriumError(RuntimeError):
    """RuntimeError regarding equilibrium."""

class DimensionError(ValueError):
    """ValueError regarding wrong dimensions."""

class UndefinedCompound(AttributeError):
    """LookupError regarding undefined compounds."""
    def __init__(self, ID): super().__init__(f"'{ID}'")

# Python's built in KeyError quotes the message, used as a by-pass for the debbuger
_KE = type('KeyError', (Exception, ), {})


# %% Biosteam Warnings

class DesignWarning(Warning):
    """Warning regarding design constraints."""
    
#%% Decorators and functions

# def _notify_error(func):
#     """Decorate class method to provide a location summary when an error occurs."""
#     if hasattr(func, '_original'):
#         func = func._original

#     def wrapper(self, *args, **kwargs):
#         try: return func(self, *args, **kwargs)
#         except Exception as e:
#             # If exception already include location, it is replaced
#             location = f'@{type(self).__name__} {self}'
#             msg = str(e).strip('\n').replace(location + ': ', '')
            
#             # Add location to message
#             if not ('@' in msg and ':\n' in msg):
#                 msg = location + f'.{func.__name__}:\n' + msg                 
            
#             # Raise exception with same traceback but new message
#             if type(e) is KeyError:
#                 raise _KE(msg).with_traceback(_sys.exc_info()[2])
#             else:
#                 raise type(e)(msg).with_traceback(_sys.exc_info()[2])
    
#     wrapper.__name__ = func.__name__
#     wrapper.__doc__ = func.__doc__
#     wrapper._original = func
#     wrapper.__annotations__ = func.__annotations__
#     return wrapper

def _try_method(method):
    try: return method()
    except Exception as e:
        # If exception already include location, it is replaced
        try:
            self = method.__self__
        except:
            raise e
        location = f'@{type(self).__name__} {self}'
        msg = str(e).strip('\n').replace(location + ': ', '')
        
        # Add location to message
        if not ('@' in msg and ':\n' in msg):
            msg = location + f'.{method.__name__}:\n' + msg                 
        
        # Raise exception with same traceback but new message
        if type(e) is KeyError:
            raise _KE(msg).with_traceback(_sys.exc_info()[2])
        elif type(e) is UndefinedCompound:
             raise e
        else:
            raise type(e)(msg).with_traceback(_sys.exc_info()[2])




# -*- coding: utf-8 -*-
"""
Created on Wed Mar  6 16:53:04 2019

@author: yoelr
"""
from . import _parameter
from . import _block
from . import _state
from . import _model
from ._parameter import *
from ._block import *
from ._state import *
from ._model import *
from . import tools

__all__ = ['tools']
__all__.extend(_parameter.__all__)
__all__.extend(_block.__all__)
__all__.extend(_state.__all__)
__all__.extend(_model.__all__)


# -*- coding: utf-8 -*-
"""
Created on Sat Aug 18 13:26:29 2018

@author: yoelr
"""
from scipy import integrate
from math import log
from .. import _Q
from .._utils import DisplayUnits

__all__ = ('Compound',)

_R = 8.3144598  # Universal gas constant (J/mol-K)

# %% Integrals methods

# Methods taking the integral of a function
def _integral_rigorous(func, Xi, Xf):
    """Take the full integral."""
    return integrate.quad(func, Xi, Xf)[0]

def _integral_initial(func, Xi, Xf):
    """Take the integral with a constant derivative at the initial value."""
    return func(Xf)*(Xf-Xi)

def _integral_final(func, Xi, Xf):
    """Take the integral with a constant derivative at the final value."""
    return func(Xf)*(Xf-Xi)

def _integral_average(func, Xi, Xf):
    """Take the integral with a constant derivative at the average value."""
    return func((Xi+Xf)/2)*(Xf-Xi)


# %% Compound class

class Compound:
    """Abstract class, Compound, for making objects that contain pure component thermodynamic and transport properties. The enthalpy property, H, ignores excess ethalpies and calculates based on latent heats and the heat capacity at a constant pressure of 101325 Pa. It is suitable for solids and liquids which are weak functions of pressure

    **Parameters**

         **IDs:** [str] A unique identification

         **T:** [float] Temperature (K)

         **P:** [float] Pressure (Pa)

         **phase:** [str] 'l'(iquid) or 'g'(as)           

    **Class Attributes**

         **phase_ref** = 'l': Reference phase

         **H_ref** = 0: Reference enthalpy (kJ/hr)

         **T_ref** = 298.15: Reference temperature (K)
         
         **P_ref** = 101325: Reference pressure (Pa)

    """
    #: [DisplayUnits] Units of measure for IPython display
    display_units = DisplayUnits(T='K', P='Pa')
    
    #: [dict] Units of measure for material properties (class attribute). 
    units = dict(MW='g/mol',
                 T='K',
                 P='Pa',
                 H='J/mol',
                 S='J/mol',
                 G='J/mol',
                 U='J/mol',
                 A='J/mol',
                 Hf='J/mol',
                 Vm='m^3/mol',
                 Cpm='J/mol',
                 Cp='J/g/K',
                 rho='kg/m^3',
                 rhom='mol/m^3',
                 nu='m^2/s',
                 mu='Pa*s',
                 sigma='N/m',
                 k='W/m/K',
                 alpha='m^2/s')
    
    phase_ref = 'l' # Reference phase
    P_ref = 101325  # Reference pressure (Pa)
    T_ref = 298.15  # Reference temperature (K)
    S_ref = 0       # Reference entropy (kJ/kmol)
    H_ref = 0       # Reference enthalpy (kJ/kmol)
    MW = 1          # Arbitrary molecular weight (g/mol)

    #: Method for taking the integral of Cp
    _integral = staticmethod(_integral_average)

    def quantity(self, prop_ID):
        """Return a material property as a Quantity object as described in the `pint package <https://pint.readthedocs.io/en/latest/>`__ 

        **Parameters**

             **prop_ID:** *[str]* name of the property (e.g. 'mol', 'H', 'k' ...)
             
        """
        return _Q(getattr(self, prop_ID), self.units[prop_ID])

    ### Abstract methods ###
    
    def HeatCapacitySolid(self, T): return self.Cpsm
    def HeatCapacityLiquid(self, T): return self.Cplm
    def HeatCapacityGas(self, T): return self.Cpgm
    def calc_H_excess(self, T, P): return 0    
    def calc_S_excess(self, T, P): return 0

    ### Entropy methods ###
    
    def _delSs_delT(self, T0, T):
        """Change in Entropy of a solid due to change in temperature at constant pressure."""
        return self.HeatCapacitySolid((T+T0)/2)*log(T/T0)

    def _delSl_delT(self, T0, T):
        """Change in Entropy of a liquid due to change in temperature at constant pressure."""
        return self.HeatCapacityLiquid((T+T0)/2)*log(T/T0)

    def _delSg_delT(self, T0, T):
        """Change in Entropy of a gas due to change in temperature at constant pressure."""
        return self.HeatCapacityGas((T+T0)/2)*log(T/T0)

    ### Energies ###

    @property
    def H(self):
        """Enthapy (kJ/kmol) disregarding pressure and assuming the specified phase."""
        phase_ref = self.phase_ref
        phase = self.phase

        # Perform enthalpy calculations at 101325 Pa
        if phase_ref == phase:
            if phase == 'l':
                H = self._integral(self.HeatCapacityLiquid, self.T_ref, self.T)
            elif phase == 'g':
                H = self._integral(self.HeatCapacityGas, self.T_ref, self.T)
            elif phase == 's':
                H = self._integral(self.HeatCapacitySolid, self.T_ref, self.T)
        elif phase_ref == 'l' and phase == 'g':
            H = self.H_int_l_T_ref_l_to_Tb + self.Hvap_Tbm + \
                self._integral(self.HeatCapacityGas, self.Tb, self.T)
        elif phase_ref == 'g' and phase == 'l':
            H = -self.H_int_Tb_to_T_ref_g - self.Hvap_Tbm + \
                self._integral(self.HeatCapacityLiquid, self.Tb, self.T)
        elif phase_ref == 's' and phase == 'l':
            H = self.H_int_T_ref_s_to_Tm + self.Hfusm + \
                self._integral(self.HeatCapacityLiquid, self.Tm, self.T)
        elif phase_ref == 'l' and phase == 's':
            H = -self.H_int_l_Tm_to_T_ref_l - self.Hfusm + \
                self._integral(self.HeatCapacitySolid, self.Tm, self.T)
        elif phase_ref == 's' and phase == 'g':
            H = self.H_int_T_ref_s_to_Tm + self.Hfusm + self.H_int_l_Tm_to_Tb + \
                self.Hvap_Tbm + self._integral(self.HeatCapacityGas, self.Tb, self.T)
        elif phase_ref == 'g' and phase == 's':
            H = -self.H_int_Tb_to_T_ref_g - self.Hvap_Tbm - self.H_int_l_Tm_to_Tb - \
                self.Hfusm + self._integral(self.HeatCapacitySolid, self.Tm, self.T)
        return self.H_ref + H

    @property
    def S(self):
        """Entropy (kJ/kmol) assuming the specified phase."""
        S = self.S_ref
        phase = self.phase
        phase_ref = self.phase_ref

        # Add Pressure Entropy
        if self.phase == 'g':
            S += -_R*log(self.P/self.P_ref)

        # Add Temperature Entropy
        if phase == phase_ref:
            if phase == 'l':
                S += self._delSl_delT(self.T_ref, self.T)
            elif phase == 'g':
                S += self._delSg_delT(self.T_ref, self.T)
            elif phase == 's':
                S += self._delSs_delT(self.T_ref, self.T)
        elif phase_ref == 'l' and phase == 'g':
            S += self.S_int_l_T_ref_l_to_Tb + self.Hvap_Tbm / \
                self.Tb + self._delSg_delT(self.Tb, self.T)
        elif phase_ref == 'g' and phase == 'l':
            S += - self.S_int_Tb_to_T_ref_g - self.Hvapm / \
                self.Tb + self._delSl_delT(self.Tb, self.T)
        elif phase_ref == 's' and phase == 'l':
            S += self.S_int_T_ref_s_to_Tm + self.Hfusm / \
                self.Tm + self._delSl_delT(self.Tm, self.T)
        elif self.phase_ref == 'l' and phase == 's':
            S += - self.S_int_l_Tm_to_T_ref_l - self.Hfusm / \
                self.Tm + self._delSs_delT(self.Tm, self.T)
        elif phase_ref == 's' and phase == 'g':
            S += self.S_int_T_ref_s_to_Tm + self.Hfusm/self.Tm + \
                self.S_int_l_Tm_to_Tb + self.Hvap_Tbm / \
                self.Tb + self._delSg_delT(self.Tb, self.T)
        elif phase_ref == 'g' and phase == 's':
            S += - self.S_int_Tb_to_T_ref_g - self.Hvap_Tbm/self.Tb - \
                self.S_int_l_Tm_to_Tb - self.Hfusm / \
                self.Tm + self._delSs_delT(self.Tm, self.T)
        else:
            raise Exception(f'Error in Compound object "{self.ID}" with phase "{phase}" and reference "{phase_ref}.')
        return S
    
    ### Class factory ###

    @classmethod
    def Factory(cls, ID, const_prop_molar_dct={}, state_dep_prop_molar_dct={}):
        """Make new child class with constant and temperature dependent properties.

        **Parameters**

             **ID:** *[str]* ID of new Compound object

             **const_prop_molar_dct:** *[dict]* constant properties to set to new object

             **state_dep_prop_molar_dct:** *[dict]* methods that will turn into Python properties for the new object

        """
        child = type(ID, (cls,), {})
        
        # Set contant attributes
        for attr, const in const_prop_molar_dct.items():
            setattr(child, attr, const)
        
        # Set state dependent property
        for attr, fun in state_dep_prop_molar_dct.items():            
            setattr(child, attr, property(fun))

        # Replace the heat capacity objects for constants
        Cpsm = const_prop_molar_dct.get('Cpsm')
        Cplm = const_prop_molar_dct.get('Cplm')
        Cpgm = const_prop_molar_dct.get('Cpgm')
        if Cpsm:
            setattr(child, 'HeatCapacitySolid', lambda self, T: Cpsm)
        if Cplm:
            setattr(child, 'HeatCapacityLiquid', lambda self, T: Cplm)
        if Cpgm:
            setattr(child, 'HeatCapacityGas', lambda self, T: Cpgm)

        # Replace the heat capacity objects for functions
        Cpsm = const_prop_molar_dct.get('Cpsm')
        Cplm = const_prop_molar_dct.get('Cplm')
        Cpgm = const_prop_molar_dct.get('Cpgm')
        if Cpsm:
            setattr(child, 'HeatCapacitySolid', property(Cpsm))
        if Cplm:
            setattr(child, 'HeatCapacityLiquid', property(Cplm))
        if Cpgm:
            setattr(child, 'HeatCapacityGas', property(Cpgm))
        
        # Replace the heat capacity objects if just one value
        Cpm = const_prop_molar_dct.get('Cpm')
        if Cpm:
            # Replace the heat capacity objects
            setattr(child, 'HeatCapacitySolid', lambda T: const)
            setattr(child, 'HeatCapacityLiquid', lambda T: const)
            setattr(child, 'HeatCapacityGas', lambda T: const)
        
        return child

    @classmethod
    def set_integral_type(cls, integral_type):
        """Change the enthalpy and entropy property calculator to different levels of rigour.

        **Parameters**

             int_type: [str] One of the following:
                      * 'rigorous': Compute integral by full integration
                      * 'average': Compute integral at an average function value
                      * 'constant': Compute integral using current function value
        """
        if integral_type == 'rigorous':
            cls._integral = staticmethod(_integral_rigorous)
        elif integral_type == 'average':
            cls._integral = staticmethod(_integral_average)
        elif integral_type == 'final':
            cls._integral = staticmethod(_integral_final)
        elif integral_type == 'initial':
            cls._integral = staticmethod(_integral_final)
        else:
            raise ValueError("Must pass one of the following types of Cp integration: 'rigorous', 'average', 'initial', or 'final'")
        cls._integral_type = integral_type

    # Representation
    def _info(self, T, P):
        """Return string with all specifications."""
        units = self.units
        T_units, P_units = self.display_units
        T_units = T or T_units
        P_units = P or P_units
        
        # First line
        info = f"{type(self).__name__}: {self.ID}\n"
        
        # Second line (thermo)
        T = _Q(self.T, units['T']).to(T_units).magnitude
        P = _Q(self.P, units['P']).to(P_units).magnitude
        info += f" phase: '{self.phase}', T: {T:.5g} {T_units}, P: {P:.6g} {P_units}"
        
        return info

    def __str__(self):
        return self.ID

    def show(self, T=None, P=None):
        """print information on self"""
        print(self._info(T, P))
    _ipython_display_ = show

    def __repr__(self):
        return f'<{type(self).__name__}: {self.ID}>'




# -*- coding: utf-8 -*-
"""
Created on Wed Dec  5 16:47:33 2018

This module includes classes and functions relating Stream objects.

@author: Yoel Cortes-Pena
"""
__all__ = ('MissingStream', 'Ins', 'Outs',
           'Sink', 'Source')

# # For later use in this module
# def missing_method(self, *args, **kwargs):
#     raise TypeError(f'Method not supported for {type(self).__name__} object.')


# %% Dummy Stream object

class MissingStreamType:
    """Create a MissingStream object that acts as a dummy in Ins and Outs objects until replaced by an actual Stream object."""
    _sink = None
    _source = None
    
    def __new__(cls):
        return MissingStream
    
    def __bool__(self):
        return False
    
    def __setattr__(self, key, value): pass
    
    def __getattr__(self, key):
        raise AttributeError(type(self).__name__)

    def __repr__(self):
        return f'MissingStream'
    
    def __str__(self):
        return 'missing stream'

MissingStream = object.__new__(MissingStreamType)


# %% List objects for input and output streams

class Ins(list):
    """Create a Ins object which serves as a list of input streams for a Unit object."""
    __slots__ = ('sink',)

    def __init__(self, sink, streams=()):
        self.sink = sink #: Unit object where Ins object is attached
        self._clear_sink(sink)
        super().__init__(streams)
        self._fix_sink(sink)
    
    def pop(self, index):
        s = super().pop(index)
        s._sink = None
        return s
    
    def insert(self, index, stream):
        super().insert(index, stream)
        stream._sink = self.sink
    
    def remove(self, stream):
        stream._sink = None
        super().remove(stream)
    
    def _clear_sink(self, sink):
        """Remove sink from streams."""
        for s in self:
            if s._sink is sink: s._sink = None
    
    def _fix_sink(self, sink):
        """Set sink for all streams."""
        for s in self: s._sink = sink
    
    def __setitem__(self, index, stream):
        sink = self.sink
        if isinstance(index, int):
            s_old = self[index]
            if s_old._sink is sink: s_old._sink = None
            stream._sink = sink
            super().__setitem__(index, stream)
        elif isinstance(index, slice):
            self._clear_sink(sink)
            super().__setitem__(index, stream)
            self._fix_sink(sink)
        else:
            raise TypeError(f'Only intergers and slices are valid indices for {type(self).__name__} objects')
        
           
    def clear(self):
        self._clear_sink(self.sink)
        super().clear()
    
    def extend(self, iterable):
        sink = self.sink
        # Add sink to new streams
        for s in iterable: s._sink = sink
        super().extend(iterable)
        
    def append(self, stream):
        sink = self.sink
        # Add sink to new stream
        stream._sink = sink
        super().append(stream)
    

class Outs(list):
    """Create a Outs object which serves as a list of output streams for a Unit object."""
    __slots__ = ('source',)
    
    def __init__(self, source, streams=()):
        self.source = source #: Unit object where Outs object is attached
        self._clear_source(source)
        super().__init__(streams)
        self._fix_source(source)
            
    def pop(self, index):
        s = super().pop(index)
        s._source = None
        return s
    
    def insert(self, index, stream):
        super().insert(index, stream)
        stream._source = self.source
    
    def remove(self, stream):
        stream._source = None
        super().remove(stream)    
    
    def _clear_source(self, source):
        """Remove source from streams."""
        for s in self:
            if s._source is source: s._source = None
    
    def _fix_source(self, source):
        """Set source for all streams."""
        for s in self: s._source = source
            
    def __setitem__(self, index, stream):
        source = self.source
        if isinstance(index, int):
            s_old = self[index]
            if s_old._source is source: s_old._source = None
            super().__setitem__(index, stream)
            stream._source = source
        elif isinstance(index, slice):
            self._clear_source(source)
            super().__setitem__(index, stream)
            self._fix_source(source)
        else:
            raise TypeError(f'Only intergers and slices are valid indices for {type(self).__name__} objects')
        
    def clear(self):
        self._clear_source(self.source)
        super().clear()
    
    def extend(self, iterable):
        source = self.source
        # Add source to new streams
        for s in iterable: s._source = source
        super().extend(iterable)
        
    def append(self, stream):
        source = self.source
        # Add sink to new stream
        stream._source = source
        super().append(stream)


# %% Sink and Source object for piping notation

class Sink:
    """Create a Sink object that connects a stream to a unit using piping notation:
    
    **Parameters**
    
        stream: [Stream]
        
        index: [int]
        
    **Example**
    
    First create a stream and a Mixer:
    
    .. code-block:: python
    
        >>> stream = Stream('s1')
        >>> unit = Mixer('M1')
    
    Sink objects are created using -pipe- notation:
        
    .. code-block:: python
    
        >>> stream-1
        <Sink: s1-1>
    
    Use pipe notation to create a sink and connect the stream:
    
    .. code-block:: python
    
        >>> stream-1-unit
        >>> M1.show()
        
        Mixer: M1
        ins...
        [0] Missing stream
        [1] s1
            phase: 'l', T: 298.15 K, P: 101325 Pa
            flow:  0
        outs...
        [0] d27
            phase: 'l', T: 298.15 K, P: 101325 Pa
            flow:  0
    
    """
    __slots__ = ('stream', 'index')
    def __init__(self, stream, index):
        self.stream = stream
        self.index = index

    # Forward pipping
    def __sub__(self, unit):
        unit.ins[self.index] = self.stream
        return unit
    
    # Backward pipping
    __pow__ = __sub__
    
    def __repr__(self):
        return '<' + type(self).__name__ + ': ' + self.stream.ID + '-' + str(self.index) + '>'


class Source:
    """Create a Source object that connects a stream to a unit using piping notation:
    
    **Parameters**
    
        stream: [Stream]
        
        index: [int]
        
    **Example**
    
    First create a stream and a Mixer:
    
    .. code-block:: python
    
        >>> stream = Stream('s1')
        >>> unit = Mixer('M1')
    
    Source objects are created using -pipe- notation:
        
    .. code-block:: python
    
        >>> 1**stream
        <Source: 1-s1>
    
    Use -pipe- notation to create a source and connect the stream:
    
    .. code-block:: python
    
        >>> unit**0**stream
        >>> M1.show()
        
        Mixer: M1
        ins...
        [0] Missing stream
        [1] Missing stream
        outs...
        [0] s1
            phase: 'l', T: 298.15 K, P: 101325 Pa
            flow:  0
    
    """
    __slots__ = ('stream', 'index')
    def __init__(self, stream, index):
        self.stream = stream
        self.index = index

    # Forward pipping
    def __rsub__(self, unit):
        unit.outs[self.index] = self.stream
        return unit
    
    # Backward pipping
    __rpow__ = __rsub__
    
    def __repr__(self):
        return '<' + type(self).__name__ + ': ' + str(self.index) + '-' + self.stream.ID + '>'







# -*- coding: utf-8 -*-
"""
Created on Sun Jul 21 21:30:33 2019

@author: yoelr
"""
from numpy import asarray, array
from biosteam._utils import wegstein_secant, aitken_secant

__all__ = ('BubblePoint',)

class BubblePoint:
    __slots__ = ('gamma', 'P', 'T', 'y')
    rootsolver = staticmethod(aitken_secant)
    
    def __init__(self, gamma):
        self.gamma = gamma
    
    def _T_error(self, T, z_over_P, z, VPs):
        self.y =  z_over_P * array([i(T) for i in VPs]) * self.gamma(z, T)
        return 1. - self.y.sum()
    
    def _P_error(self, P, T, z, Psat_gamma):
        self.y = z * Psat_gamma / P
        return 1. - self.y.sum()
    
    def solve_Ty(self, z, P):
        """Bubble point at given composition and pressure

        **Parameters**
        
            **z:** [array_like] Liquid phase composition.
        
            **P:** [float] Pressure (Pa).
        
        **Returns**

            **T:** [float] Bubble point temperature (K)

            **y:** [numpy ndarray] Composition of the vapor phase.

        >>> from biosteam import Species, BubblePoint, Dortmund
        >>> gamma = Dortmund(*Species('Ethanol', 'Water'))
        >>> bp = BubblePoint(gamma)
        >>> bp.solve_Ty(z=(0.6, 0.4), P=101325)
        (352.2820850833474, array([0.703, 0.297]))
        
        """
        z = asarray(z)
        self.P = P
        try:
            self.T = self.rootsolver(self._T_error, self.T, self.T+0.01, 1e-6,
                                     args=(z/P, z, [s.VaporPressure for s in self.gamma.species]))
        except:
            T = (z * [s.Tb for s in self.gamma.species]).sum()
            self.T = self.rootsolver(self._T_error, T, T+0.01, 1e-6,
                                     args=(z/P, z, [s.VaporPressure for s in self.gamma.species]))
        self.y = self.y/self.y.sum()
        return self.T, self.y
    
    def solve_Py(self, z, T):
        """Bubble point at given composition and temperature.

        **Parameters**
        
            **z:** [array_like] Liquid phase composotion.

            **T:** [float] Temperature (K).
        
        **Returns**

            **P:** [float] Bubble point pressure (Pa).

            **y:** [numpy ndarray] Vapor phase composition.

        >>> from biosteam import Species, BubblePoint, Dortmund
        >>> gamma = Dortmund(*Species('Ethanol', 'Water'))
        >>> bp = BubblePoint(gamma)
        >>> bp.solve_Py(z=(0.703, 0.297), T=352.28)
        (103494.17209657285, array([0.757, 0.243]))
        
        """
        z = asarray(z)
        Psat = array([i.VaporPressure(T) for i in self.gamma.species])
        Psat_gamma =  Psat * self.gamma(z, T)
        self.T = T
        try:
            self.P = self.rootsolver(self._P_error, self.P, self.P-1, 1e-2,
                                     args=(T, z, Psat_gamma))
        except:
            P = (z * Psat).sum()
            self.P = self.rootsolver(self._P_error, P, P-1, 1e-2,
                                     args=(T, z, Psat_gamma))
        self.y = self.y/self.y.sum()
        return self.P, self.y
    
    def __repr__(self):
        return f"<{type(self).__name__}: gamma={self.gamma}>"
    

    


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:56:55 2018

@author: yoelr
"""

__all__ = ('size_batch',)

def size_batch(v_0, tau, tau_0, N_reactors, V_wf) -> dict:
        """Solve for batch reactor volume, cycle time, and loading time
        
        **Parameters**
        
            **v_0:** Flow rate
            
            **tau:** Reaction time
            
            **tau_0:** Turnaround time
            
            **N_reactors:** Number of reactors
            
            **V_wf:** Fraction of working volume
        
        **Returns:** dict[str: float]
            * 'Reactor volume'
            * 'Cycle time'
            * 'Loading time'
        """
        V_T = v_0*(tau + tau_0)/(1-1/N_reactors) # Reacting volume
        V_i = V_T/N_reactors
        t_L = V_i/v_0
        t_B = tau + tau_0 + t_L
        V_i /= V_wf
        
        return {'Reactor volume': V_i,
                'Cycle time': t_B, 
                'Loading time': t_L}
        
        


# -*- coding: utf-8 -*-
"""
Created on Tue Sep 18 10:28:29 2018

@author: Guest Group
"""
from scipy.optimize import least_squares
from ._equilibrium import VLE, BubblePoint, DewPoint, \
                          Dortmund
from ._species import Species, WorkingSpecies
from ._stream import Stream, nonzero_species, MassFlow, \
                            mol_flow_dim, mass_flow_dim, vol_flow_dim, \
                            phases, phase_index, flow
from ._utils import fraction, tuple_array, \
                    PropertyFactory, property_array
from ._exceptions import EquilibriumError, DimensionError
from . import _Q
import numpy as np

__all__ = ('MixedStream',)


# %% Dictionaries for finding the right attribute given the phase

# Used for the show method
letter_phases = {'s': 'solid',
                 'l': 'liquid',
                 'L': 'LIQUID',
                 'g': 'vapor'}

# %% Volumetric flow property

@PropertyFactory    
def VolumetricFlow(self):
    """Volumetric flow (m^3/hr)."""
    stream, mol, phase = self.data
    if mol:
        c = self.name # c: compound
        c.T = stream.T
        c.P = stream.P
        c.phase = phase
        return (c.Vm * mol[0] * 1000.)
    else:
        return 0.

@VolumetricFlow.setter
def VolumetricFlow(self, value):
    if value:
        stream, mol, phase = self.data
        c = self.name # c: compound
        c.T = stream.T
        c.P = stream.P
        c.phase = phase
        mol[0] = value/(c.Vm * 1000)
    else:
        mol[0] = 0.

# %% MixedStream (4 phase flow)

# TODO: Add solids handling:
# -Viscosity with class attribute J, the coefficient in Mc = Mw(1 - Vs/Vm)^(1/J).
# A generalized mixture rule for estimating the viscosity of solid-liquid suspensions and mechanical properties of polyphase rocks and composite materials, https://agupubs.onlinelibrary.wiley.com/doi/epdf/10.1029/2004JB003124, suggests that J = -0.5 for real suspensions including carbon fibers

# TODO: Make MixedStream capable of handling equilibrium, and mixing with Stream


class MixedStream(Stream):
    """An extension of the :doc:`Stream` class. All inherited properties and methods are supported. Create a MixedStream object that can manage mixtures of solid ('s'), liquid ('l'), LIQUID ('L'), and vapor ('g') phases. A MixedStream object is capable of vapor-liquid equilibrium (VLE) and liquid-LIQUID equilibrium (LLE).

    **Parameters**

         **ID:** [str] ID of the stream. If ID is '', a default ID will be chosen.

         **species:** [Species] Species object for the stream. If None, assume same as class' Species object.

         **T:** [float] Temperature (K)

         **P:** [float] Pressure (Pa)

    **Examples**

         :doc:`MixedStream objects and thermodynamic equilibrium`
    """
    __slots__ = ()
    def __init__(self, ID='', species=None, T=298.15, P=101325):
        # Get species and set species information
        if species is None:
            assert self._cls_species, 'must define Stream.species first'
            self._species = self._cls_species
        elif isinstance(species, Species):
            self._species = WorkingSpecies(species)
        elif isinstance(species, WorkingSpecies):
            self._species = species
        else:
            raise TypeError(f"species must be a Species object, not '{type(species).__name__}'")
        self._source_link = self
        self._link = self._source = self._sink = self._ID = None
        self.ID = ID
        self.T = T  #: [float] Temperature (K)
        self.P = P  #: [float] Pressure (Pa)
        self._lL_split_cached = (None,)
        self._gamma = gamma = Dortmund()
        self._bubble_point = BubblePoint(gamma)
        self._dew_point = DewPoint(gamma)
        self._setflows(np.zeros((4, self._species._N), dtype=np.float64))
        self._VLE = VLE(self)

    def getflow(self, phase, *species, units='kmol/hr'):
        """Get flow rates of species in given units."""
        index = self.indices(species) if species else ...
        p = phase_index[phase]
        q = _Q(1, units)
        dim = q.dimensionality
        if dim == mol_flow_dim:
            return self._mol[p, index]*q.to('kmol/hr').magnitude
        elif dim == mass_flow_dim:
            return self._mass[p, index]*q.to('kg/hr').magnitude
        elif dim == vol_flow_dim:
            return self._vol[p, index]*q.to('m3/hr').magnitude
        else:
            raise DimensionError(f"dimensions for flow units must be in molar, mass or volumetric flow rates, not '{dim}'")
        
    def setflow(self, phase, flow=(), species=(), units='kmol/hr',
                 inplace='', **flow_pairs):
        """Set flow rates according to the species order and flow_pairs. `inplace` can be any operation that can be performed in place (e.g. +, -, *, /, |, **, etc.)."""
        if species and isinstance(species[0], str):
            species = [self._species_dict[ID] for ID in species]
        species = (*species, *flow_pairs.keys())
        flow = (*flow, *flow_pairs.values())
        p = phase_index[phase]
        index = self.indices(species) if species else ... 
        q = _Q(flow, units)
        dim = q.dimensionality
        if dim == mol_flow_dim:
            exec(f"self._mol[p, index] {inplace}= q.to('kmol/hr').magnitude", locals())
        elif dim == mass_flow_dim:
            exec(f"self._mass[p, index] {inplace}= q.to('kg/hr').magnitude", locals())
        elif dim == vol_flow_dim:
            exec(f"self._vol[p, index] {inplace}= q.to('m3/hr').magnitude", locals())
        else:
            raise DimensionError(f"dimensions for flow units must be in molar, mass or volumetric flow rates, not '{dim}'")

    def _setflows(self, mol):
        """ Set molar, mass and volumetric flow rates for all phases.
        
        **Parameters**
        
            **mol_array:** Array of molar flow rates (kmol/hr) with species by column and phases ('s', 'l', 'L', 'g') by row.
        
        """
        species = self._species
        MW = species._MW
        cmps = species._compounds
        index = species._index
        massflows = [] # Mass flow rates    
        volflows = [] # Volumetric flow rates    
        for phase, m in zip(phases, mol):
            vol = []
            mass = []
            for i in index:
                s = cmps[i]
                mi = m[i:i+1]
                mass.append(MassFlow(s.ID, (mi, MW[i])))
                vol.append(VolumetricFlow(s, (self, mi, phase.lower())))
            massflows.append(mass)
            volflows.append(vol)
        self._mol = mol
        self._mass = property_array(massflows)
        self._vol = property_array(volflows)

    @property
    def link(self):
        return None
    @link.setter
    def link(self, stream):
        raise TypeError(f'{type(self).__name__} objects do not support linking')

    @property
    def phase(self):
        """String denoting all phases present"""
        phase = ''
        arr = self._mol != 0
        if arr[0].any(): phase += 's'
        if arr[1].any(): phase += 'l'
        if arr[2].any(): phase += 'L'
        if arr[3].any(): phase += 'g'
        return phase
    @phase.setter
    def phase(self, phase):
        if phase not in phases:
            raise ValueError("phase must be one of the following: 's', 'l', 'L' or 'g'")
        mol = self._mol
        val = mol.sum(0)
        mol[:] = 0
        mol[phase_index[phase]] = val

    ### Solids ###

    # Molar
    @flow
    def solid_mol(self):
        """solid phase molar flow rates (kmol/hr)."""
        return self._mol[0]
    @property
    def solid_molfrac(self):
        """solid phase molar fractions."""
        return fraction(self._mol[0]).view(tuple_array)
    @property
    def solid_molnet(self):
        """solid phase net molar flow rate (kg/hr)."""
        return self._mol[0].sum()

    # Mass
    @flow
    def solid_mass(self):
        """solid phase mass flow rates (kg/hr)."""
        return self._mass[0]
    @property
    def solid_massfrac(self):
        """solid phase mass fractions."""
        return fraction(self._mass[0]).view(tuple_array)
    @property
    def solid_massnet(self):
        """solid phase net mass flow rate (kg/hr)."""
        return self._mass[0].sum()

    # Volumetric
    @flow
    def solid_vol(self):
        """solid phase volumetric flow rates (m3/hr)."""
        return self._vol[0]
    @property
    def solid_volfrac(self):
        """solid phase volumetric fractions."""
        return fraction(self._vol[0]).view(tuple_array)
    @property
    def solid_volnet(self):
        """solid phase net volumetric flow rate (m3/hr)."""
        return self._vol[0].sum()

    ### Liquids ###

    # Molar
    @flow
    def liquid_mol(self):
        """liquid phase molar flow rates (kmol/hr)."""
        return self._mol[1]
    @property
    def liquid_molfrac(self):
        """liquid phase molar fractions."""
        return fraction(self._mol[1]).view(tuple_array)
    @property
    def liquid_molnet(self):
        """liquid phase net molar flow rate (kg/hr)."""
        return self._mol[1].sum()

    # Mass
    @flow
    def liquid_mass(self):
        """liquid phase mass flow rates (kg/hr)."""
        return self._mass[1]
    @property
    def liquid_massfrac(self):
        """liquid phase mass fractions."""
        return fraction(self._mass[1]).view(tuple_array)
    @property
    def liquid_massnet(self):
        """liquid phase net mass flow rate (kg/hr)."""
        return self._mass[1].sum()

    # Volumetric
    @flow
    def liquid_vol(self):
        """liquid phase volumetric flow rates (m3/hr)."""
        return self._vol[1]
    @property
    def liquid_volfrac(self):
        """liquid phase volumetric fractions."""
        return fraction(self._vol[1]).view(tuple_array)
    @property
    def liquid_volnet(self):
        """liquid phase net volumetric flow rate (m3/hr)."""
        return self._vol[1].sum()

    ### LIQUID ###

    # Molar
    @flow
    def LIQUID_mol(self):
        """LIQUID phase molar flow rates (kmol/hr)."""
        return self._mol[2]
    @property
    def LIQUID_molfrac(self):
        """LIQUID phase molar fractions."""
        return fraction(self._mol[2]).view(tuple_array)
    @property
    def LIQUID_molnet(self):
        """LIQUID phase net molar flow rate (kg/hr)."""
        return self._mol[2].sum()

    # Mass
    @flow
    def LIQUID_mass(self):
        """LIQUID phase mass flow rates (kg/hr)."""
        return self._mass[2]
    @property
    def LIQUID_massfrac(self):
        """LIQUID phase mass fractions."""
        return fraction(self._mass[2]).view(tuple_array)
    @property
    def LIQUID_massnet(self):
        """LIQUID phase net mass flow rate (kg/hr)."""
        return self._mass[2].sum()

    # Volumetric
    @flow
    def LIQUID_vol(self):
        """LIQUID phase volumetric flow rates (m3/hr)."""
        return self._vol[2]
    @property
    def LIQUID_volfrac(self):
        """LIQUID phase volumetric fractions."""
        return fraction(self._vol[2]).view(tuple_array)
    @property
    def LIQUID_volnet(self):
        """LIQUID phase net volumetric flow rate (m3/hr)."""
        return self._vol[2].sum()

    ### Vapor ###

    # Molar
    @flow
    def vapor_mol(self):
        """vapor phase molar flow rates (kmol/hr)."""
        return self._mol[3]
    @property
    def vapor_molfrac(self):
        """vapor phase molar fractions."""
        return fraction(self._mol[3]).view(tuple_array)
    @property
    def vapor_molnet(self):
        """vapor phase net molar flow rate (kg/hr)."""
        return self._mol[3].sum()

    # Mass
    @flow
    def vapor_mass(self):
        """vapor phase mass flow rates (kg/hr)."""
        return self._mass[3]
    @property
    def vapor_massfrac(self):
        """vapor phase mass fractions."""
        return fraction(self._mass[3]).view(tuple_array)
    @property
    def vapor_massnet(self):
        """vapor phase net mass flow rate (kg/hr)."""
        return self._mass[3].sum()

    # Volumetric
    @flow
    def vapor_vol(self):
        """vapor phase volumetric flow rates (m3/hr)."""
        return self._vol[3]
    @property
    def vapor_volfrac(self):
        """vapor phase volumetric fractions."""
        return fraction(self._vol[3]).view(tuple_array)
    @property
    def vapor_volnet(self):
        """vapor phase net volumetric flow rate (m3/hr)."""
        return self._vol[3].sum()
    
    ### Overall ###

    # Flow rates
    @property
    def mol(self):
        """Molar flow rates (kmol/hr)."""
        return self._mol.sum(0).view(tuple_array)
    @property
    def mass(self):
        """Mass flow rates (kmol/hr)."""
        return (self._mol.sum(0) * self._species._MW).view(tuple_array)
    @property
    def vol(self):
        """Volumetric flow rates (kmol/hr)."""
        return self._vol.sum(0).view(tuple_array)

    # Fractions
    @property
    def molfrac(self):
        """Molar fractions."""
        return fraction(self._mol.sum(0)).view(tuple_array)
    @property
    def massfrac(self):
        """Mass fractions."""
        return fraction(self._mol.sum(0) * self._species._MW).view(tuple_array)
    @property
    def volfrac(self):
        """Volumetric fractions."""
        return fraction(self._vol.sum(0)).view(tuple_array)

    # Net flow rates
    @property
    def molnet(self):
        """Mol net flow rate (kmol/hr)."""
        return self._mol.sum()
    @property
    def massnet(self):
        """Mass net flow rate (kmol/hr)."""
        return (self._mol.sum(0)*self._species._MW).sum()
    @property
    def volnet(self):
        """Volumetric net flow rate (kmol/hr)."""
        return self._vol.sum()

    # Other properties
    @property
    def mu(self):
        # Katti, P.K.; Chaudhri, M.M. (1964). "Viscosities of Binary Mixtures of Benzyl Acetate with Dioxane, Aniline, and m-Cresol". Journal of Chemical and Engineering Data. 9 (1964): 442–443.
        # molnet = self.molnet
        # if self.molnet == 0: return 0
        # N = self._N
        # mol = self._mol
        # mus = np.zeros(4, N)
        # Vms = np.zeros(4, N)
        # props = self._species._props
        # T = self._T
        # P = self._P
        # mus[0] = props('mu', mol[0], T, P, 's')
        # Vms[0] = props('Vm', mol[0], T, P, 's')
        # mus[1] = props('mu', mol[1], T, P, 'l')
        # Vms[1] = props('Vm', mol[1], T, P, 'l')
        # mus[2] = props('mu', mol[2], T, P, 'l')
        # Vms[2] = props('Vm', mol[2], T, P, 'l')
        # mus[3] = props('mu', mol[3], T, P, 'g')
        # Vms[3] = props('Vm', mol[3], T, P, 'g')
        # pos = np.where(mol != 0)
        # mol = mol[pos]
        # molfrac = mol/molnet
        # return np.exp((molfrac*np.log(mus[pos]*Vms[pos])).sum())/self.Vm
        return self._species._mixedprop('mu', self._mol, self.T, self.P)
    mu.__doc__ = Stream.mu.__doc__
    
    ### Energy flows ###

    @property
    def H(self):
        """Enthalpy flow rate as a function of T, excluding formation energies (kJ/hr).

        >>> s1.H # The stream is at the reference state
        0.0
        >>> s1.H = 1000
        >>> s1.T
        304.80600901692236
        >>> s1.H
        1002.1604769775776

        .. Note:: The solver reaches a level of tolerance and the new temperature is an approximation.
        """
        return self._species._mixedpropflow('H', self._mol, self.T, self.P)
    H = H.setter(Stream.H.fset)
        
    @property
    def S(self):
        """Entropy flow rate as a function of T and P, excluding formation energies (kJ/hr).

        >>> s1.S # The stream is at the reference state
        0.0
        >>> s1.T = 320
        >>> s1.S
        10.642980522582695
        """
        return self._species._mixedpropflow('S', self._mol, self.T, self.P)

    @property
    def C(self):
        """Heat capacity flow rate as a function of T (kJ/K/hr).

        >>> s2.C
        262.74816631655267
        """
        return self._species._mixedpropflow('Cpm', self._mol, self.T, self.P)

    @property
    def Vm(self):
        """Molar volume as a function of T and P (m^3/mol).

        >>> s1.Vm
        1.8069039870122814e-05
        """
        return self._species._mixedprop('Vm', self._mol, self.T, self.P)

    @property
    def k(self):
        """Thermal conductivity as a function of T and P (W/m/k).

        >>> s1.k
        0.5942044328004411
        """
        return self._species._mixedprop('k', self._mol, self.T, self.P)

    @property
    def alpha(self):
        """Thermal diffusivity as a function of T and P (m^2/s).

        >>> s1.alpha
        1.4255801521655763e-07
        """
        return self._species._mixedprop('alpha', self._mol, self.T, self.P)

    @property
    def sigma(self):
        """Surface tension as a function of T (N/m).

        >>> s1.sigma
        0.07205503890847455
        """
        return self._species._mixedprop('sigma', self._mol, self.T, self.P)

    @property
    def Pr(self):
        """Prandtl number as a function of T and P (non-dimensional).

        >>> s1.Pr
        6.421553249380879
        """
        return self._species._mixedprop('Pr', self._mol, self.T, self.P)
    
    ### Specifications ###

    def copylike(self, stream):
        """Copy flow rates, T, P, and phase of stream to self."""
        self.copyflow(stream)
        self.P = stream.P
        self.T = stream.T

    def copyflow(self, stream, species=None, remove=False, exclude=False):
        """Copy flow rates of stream to self.
        
        **Parameters**
        
            **stream:** [Stream] Flow rates will be copied from here.
            
            **species:** iterable[str] Species IDs. Defaults to all species.
            
            **remove:** [bool] If True, copied species will be removed from `stream`.
            
            **exclude:** [bool] If True, exclude `species` when copying.
            
        .. Note::
            
            Species flow rates that are not copied will be set to zero.
        
        """
        assert self._species is stream._species, ('species must be the same to '
                                                  'copy stream specifications')
        if isinstance(stream, MixedStream):
            if species is None:
                self._mol[:] = stream._mol
                if remove: stream._mol[:] = 0
            else:
                indices = self.indices(species)
                if exclude:
                    self._mol[:] = stream._mol
                    self._mol[indices] = 0
                    if remove:
                        mol = stream._mol
                        mol[:], mol[indices] = 0, mol[indices]
                else:
                    mol[:] = stream._mol[:, indices]
                    if remove: stream.mol[:, indices] = 0
        elif isinstance(stream, Stream):
            if species is None:
                self._mol[:] = 0
                self._mol[phase_index[stream._phase]] = stream.mol
                if remove: stream.mol[:] = 0
            else:
                indices = self.indices(species)
                if exclude:
                    p = phase_index[stream._phase]
                    self._mol[p, :] = stream._mol
                    self._mol[p, indices] = 0
                    if remove:
                        mol = stream._mol
                        mol[:], mol[indices] = 0, mol[indices]
                else:
                    self._mol[:] = 0
                    self._mol[phase_index[stream._phase], indices] = stream._mol[indices]
                    if remove: stream.mol[indices] = 0
        else:
            raise TypeError('argument must be a Stream object')
            
    ### Equilibrium ###

    def disable_phases(self, phase):
        """Cast stream into a Stream object.
        
        **Parameters**
        
            **phase:** {'s', 'l', 'g'} desired phase of stream
            
        """
        self.__class__ = Stream
        self.__init__(ID=self._ID, flow=self.mol, T=self.T, P=self.P, phase=phase)

    def enable_phases(self):
        """Cast stream into a MixedStream object."""

    @property
    def VLE(self):
        """A callable VLE object for vapor-liquid equilibrium.

        **Parameters**
        
            **Specify two:**
                * **P:** [float] Operating pressure (Pa)
                * **Q:** [float] Energy input (kJ/hr)
                * **T:** [float] Operating temperature (K)
                * **V:** [float] Molar vapor fraction
                * **x:** [numpy array] Molar composition of liquid (for binary mixture)
                * **y:** [numpy array] Molar composition of vapor (for binary mixture)

            **Optional:**
            
                **species_IDs:** [tuple] IDs of species in equilibrium.
                     
                **LNK:** tuple[str] Light non-keys that remain as a vapor.
        
                **HNK:** tuple[str] Heavy non-keys that remain as a liquid.

        .. Note:
           LNK and HNK are not taken into account for equilibrium. Parameters not specified are None by default.

        """
        return self._VLE

    def LLE(self, species_IDs=None, split=None, lNK=(), LNK=(),
            solvents=(), solvent_split=(),
            P=None, T=None, Q=0):
        """Perform LIQUID-liquid equilibrium.

        **Optional Parameters**

            **species_IDs:** *tuple[str]* IDs of equilibrium species
            
            **split:** *tuple[float]* Initial guess split fractions of each specie to the 'liquid'

            **lNK:** *tuple[str]* Species assumed to completely remain in the 'liquid' phase.

            **LNK:** *tuple[str]* Species assumed to completely remain in the 'LIQUID' phase.

            **solvents:** *tuple[str]* Species corresponding to specified solvent_split

            **solvent_split:** *tuple[float]* Split fractions of each specie to the 'liquid' phase.                
            
            **Q:** *[float]* Energy input (kJ/hr)
            
            **T:** *[float]* Operating temperature (K)
            
            **P:** *[float]* Operating pressure (Pa)    

        .. Note:
           lNK and LNK are not taken into account for equilibrium. Assumes constant pressure and temperatrue if none are provided.

        """
        ### Set Up ###

        # Get molar flow rates
        liquid_mol = self.liquid_mol
        LIQUID_mol = self.LIQUID_mol
        all_mol = liquid_mol + LIQUID_mol
        notzero = all_mol > 0

        # Reused attributes
        _species = self._species
        sp_dict = _species.__dict__
        sp_index = _species._indexdct

        # Set up indices for both equilibrium and non-equilibrium species
        if species_IDs:
            index = [sp_index[specie] for specie in species_IDs]
            species = [sp_dict[ID] for ID in species_IDs]
        else:
            index = _species._equilibrium_indices(notzero)
            cmps = _species._compounds
            species = [cmps[i] for i in index]
        
        for s, i in zip(species, index):
            ID = s.ID
            if (ID in solvents) or (ID in lNK) or (ID in LNK):
                species.remove(s)
                index.remove(i)

        lNK_index = [sp_index[s] for s in lNK]
        LNK_index = [sp_index[s] for s in LNK]
        solvent_index = [sp_index[s] for s in solvents]

        # Get min and max splits
        Nspecies = len(species)
        min_ = np.zeros(Nspecies)

        # Set up activity coefficients
        solvent_species = [sp_dict[ID] for ID in solvents]
        activity_species = species + solvent_species
        self._gamma.species = species
        gamma = self._gamma

        # Make sure splits are given as arrays
        if split is None:
            split = self._lL_split_cached[1] if self._lL_split_cached[0] == species else None
            if split is None:
                split_dipoles = [s.dipole for s in species]
                solvent_dipoles = [s.dipole for s in solvent_species]
                dipoles = split_dipoles + solvent_dipoles
                split_dipoles = np.array(split_dipoles)
                try:
                    split = split_dipoles/max(dipoles)
                except TypeError as TE:
                    missing_dipoles = []
                    for i, is_missing in enumerate(split_dipoles==None):
                        if is_missing:
                            missing_dipoles.append(activity_species[i])
                    raise EquilibriumError(f'cannot make estimate for split due to missing dipole values for species: {missing_dipoles}')
                split = np.array(split)
            solvent_split = np.array(solvent_split)

        # Equilibrium by constant temperature or with duty
        if T:
            self.T = T
        elif Q:
            self.H += Q
            T = self.T
        else:
            T = self.T
        
        if P:
            self.P = P
        else:
            P = self.P

        # Equilibrium species
        mol = all_mol[index]  # Solute species flow rates
        Kmol = all_mol[solvent_index]  # Solvent species flow rates
        Kguess = Kmol * solvent_split  # Solvent species flow rates to 'l' phase
        molKmol = np.concatenate((mol, Kmol))  # Both

        # Non-equilibrium species
        lNK_mol = all_mol[lNK_index]
        LNK_mol = all_mol[LNK_index]

        ### Solve ###
        
        # Error function
        def guess_error(l_guess):
            # Error function for constant T and P, where l represents 'l' liquid flow rates
            l_guess = np.concatenate((l_guess, Kguess))
            L_guess = molKmol - l_guess
            x1_guess = l_guess/sum(l_guess)
            x2_guess = L_guess/sum(L_guess)
            x1_gamma = gamma(x1_guess, T)
            x2_gamma = gamma(x2_guess, T)
            err = (x1_guess * x1_gamma - x2_guess * x2_gamma)
            return abs(err)[0:Nspecies]

        # Guestimates
        pos = mol == 0
        mol[pos] = 10**-20  # Give it a little to not hit bounds for least squares
        l_guess = split * mol  # Initial guess

        # Solve
        sol = least_squares(guess_error, l_guess, bounds=(min_, mol))
        l_guess = sol.x
        l_guess[pos] = 0
        split = l_guess/mol
        
        # Make sure two phases are given
        xl = l_guess/sum(l_guess)
        L_guess = mol-l_guess
        xL = L_guess/sum(L_guess)
        if (xl - xL < 0.1).all():
            raise EquilibriumError('could not solve equilibrium, please input better split guesses')

        # Set results
        self._lL_split_cached = (species, split)
        liquid_mol[index] = l_guess
        liquid_mol[lNK_index] = lNK_mol
        liquid_mol[LNK_index] = 0
        liquid_mol[solvent_index] = Kmol * solvent_split
        LIQUID_mol[index] = mol-l_guess
        LIQUID_mol[lNK_index] = 0
        LIQUID_mol[LNK_index] = LNK_mol
        LIQUID_mol[solvent_index] = Kmol - liquid_mol[solvent_index]

    def _info(self, T, P, flow, fraction):
        """Return string with all specifications."""
        units = self.units
        basic_info = self._info_header() + '\n'
        if hasattr(self, '_mol'):
            nonzero, species = self.nonzero_species
        else:
            return basic_info + f' link: {self._link}'
        T_units, P_units, flow_units, fraction = [(i if i is not None else j) for i, j in
                                                  zip((T, P, flow, fraction), self.display_units)]
        phases = self.phase
        basic_info += self._info_phaseTP(phases, T_units, P_units)
        len_ = len(nonzero)
        if len_ == 0: return basic_info + ' flow: 0'
        
        # Get flow attribute name
        flow_dim = _Q(0, flow_units).dimensionality
        if flow_dim == mol_flow_dim:
            flow_type = 'mol'
        elif flow_dim == mass_flow_dim:
            flow_type = 'mass'
        elif flow_dim == vol_flow_dim:
            flow_type = 'vol'
        else:
            raise DimensionError(f"dimensions for flow units must be in molar, mass or volumetric flow rates, not '{flow_dim}'")
        
        def flow_rate_skeleton(phase, flow_type, flow_units, fraction, maxlen):
            phase_full_name = letter_phases[phase]
            beginning = f' {phase_full_name}: '
            new_line_spaces = len(beginning) * ' '

            if fraction:
                flowname = flow_type + 'net'
                flownet = getattr(self, phase_full_name + '_' + flowname)
                end = ('\n' + new_line_spaces + 'net' +(maxlen-3)*' ' + '  ' +
                       f'{flownet:.3g} {flow_units}')
            else:
                end = ''

            return beginning, new_line_spaces, end

        # Length of species column
        all_lengths = [len(sp) for sp in species]
        maxlen = max(all_lengths + [7])  # include length of the word 'species'

        # Set up flow rate information for all phases
        first_phase = True
        phases_flowrates_info = ''
        for phase in phases:
            # Get flow rates of phase in nice structure
            flow_attr = letter_phases[phase] + '_' + flow_type
            if fraction: flow_attr += 'frac'
            flow = getattr(self, flow_attr)
            nonzero, species = nonzero_species(self._species, flow)
            if fraction:
                flow_nonzero = flow[nonzero]
            else:
                flow_nonzero = _Q(flow[nonzero], units[flow_type]).to(flow_units).magnitude
            
            # Get basic structure for phase flow rates
            beginning, new_line_spaces, end = flow_rate_skeleton(phase, flow_type, flow_units, fraction, maxlen)

            # Make up for soild and vapor phase having 5 letters
            # (both liquid and LIQUID have 6 letters)
            if phase in 'sg':
                beginning += ' '
                new_line_spaces += ' '
                if fraction: end = '\n ' + end[1:]

            # Set flow rates info
            flowrates = ''
            l = len(nonzero)
            lengths = [len(sp) for sp in species]
            for i in range(l-1):
                spaces = ' ' * (maxlen - lengths[i])
                flowrates += f'{species[i]} ' + spaces + \
                    f' {flow_nonzero[i]:.3g}\n' + new_line_spaces
            spaces = ' ' * (maxlen - lengths[l-1])
            flowrates += (f'{species[l-1]} ' + spaces
                          + f' {flow_nonzero[l-1]:.3g}')

            # Add header to flow rates
            if first_phase:
                spaces = ' ' * (maxlen - 7)
                beginning = (f'{new_line_spaces}species{spaces}  {flow_units}\n'
                             + beginning)

            # Put it together
            phases_flowrates_info += beginning + flowrates + end + '\n\n'
            first_phase = False
            
        return basic_info + phases_flowrates_info[:-2]


Stream.VLE.__doc__ = MixedStream.VLE.__doc__
Stream.LLE.__doc__ = MixedStream.LLE.__doc__
MixedStream.mu.__doc__ = Stream.mu.__doc__



# -*- coding: utf-8 -*-
"""
Created on Sat Aug 18 13:42:33 2018

@author: yoelr
"""
from . import compounds as cmps
from ._exceptions import UndefinedCompound
import numpy as np

inf = np.inf
__all__ = ('Species', 'WorkingSpecies')


# TODO: Fix how Dortmund groups are selected in thermo.Chemical. Glycerol defaults to wrong value.

# %% Managing ChEDL thermodynamic properties
_setattr = object.__setattr__

class Species:
    """Create Species object that contains Compound objects as attributes.

    **Parameters**

        ***IDs:** [str] Strings should be one of the following [-]:
                  * Name, in IUPAC form or common form or a synonym registered in PubChem
                  * InChI name, prefixed by 'InChI=1S/' or 'InChI=1/'
                  * InChI key, prefixed by 'InChIKey='
                  * PubChem CID, prefixed by 'PubChem='
                  * SMILES (prefix with 'SMILES=' to ensure smiles parsing)
                  * CAS number
         
        **Optional**
        
        **cls:** [type Compound] The type of objects that will be created and stored as attributes. Defaults to biosteam.Chemical
        
    """
    @classmethod
    def tospecies(cls, compounds):
        """Return a Species object from an iterable of Compound objects."""
        self = cls.__new__(cls)
        for c in compounds: 
            if isinstance(c, cmps.Compound):
                _setattr(self, c.ID, c)
            else:
                raise TypeError('can only set Compound objects as attributes')
        return self
    
    def __init__(self, *IDs, cls=None):
        # Sphinx docs look ugly if a class is in the default argument
        if not IDs: return
        if cls is None: cls = cmps.Chemical
        
        # Set Compound object attributes
        for n in IDs: _setattr(self, n, cls(n))
    
    @property
    def IDs(self):
        return tuple(self.__dict__)
    
    def __len__(self):
        return len(self.__dict__)
    
    def __contains__(self, compound):
        return compound in self.__dict__.values()
    
    def __iter__(self):
        yield from self.__dict__.values()
    
    def __setattr__(self, ID, compound):
        if isinstance(compound, cmps.Compound):
            compound.ID = ID
            _setattr(self, ID, compound)
        elif ID == 'IDs':
            _setattr(self, '__dict__', {i: getattr(self, i) for i in compound})
        else:
            raise TypeError('can only set Compound objects as attributes')
      
    def extend(self, compounds):
        """Extend species object with more compounds."""
        for c in compounds: 
            if isinstance(c, cmps.Compound): _setattr(self, c.ID, c)
            else: raise ValueError('compounds must be an iterable of Compound objects')
            
    def getprops(self, IDs, prop_ID, T, P, phase):
        """Return list of the desired property, prop_ID, for each compound in IDz."""
        props = []; getattr_ = getattr
        for ID in IDs:
            s = getattr_(self, ID)
            s.T = T
            s.P = P
            s.phase = phase
            props.append(getattr_(s, prop_ID))
        return props
    
    def setprops(self, IDs, prop_ID, new_values):
        """Set new values to compound property, prop_ID, for given IDs."""
        getattr_ = getattr
        for ID in IDs: 
            _setattr(getattr_(self, ID), prop_ID, new_values)
    
    def __repr__(self):
        return f'<{type(self).__name__}: {", ".join(self.IDs)}>'


class WorkingSpecies:
    """Cast a Species object as a WorkingSpecies object that represents the working species of a process."""
    
    @classmethod
    def subgroup(cls, working_species, IDs):
        assert isinstance(working_species, cls), f"working_species must be a '{cls.__name__}' object"
        self = Species()
        for i in IDs: setattr(self, i, getattr(working_species, i))
        WorkingSpecies(self)
        for i in self.IDs:
            for j in working_species.get_synonyms(i):
                try: self.set_synonym(i, j)
                except: pass
        return self
    
    def __new__(cls, self):
        if self.__class__ is cls: return self
        _setattr(self, '__class__', cls)
        me = self.__dict__
        tup = tuple
        cmps = tup(me.values())
        IDs = tup(i for i in me)
        CAS = tup(i.CAS for i in cmps)
        N = len(IDs)
        index = tup(range(N))
        for i in cmps: me[i.CAS] = i
        me['_indexdct'] = dict((*zip(CAS, index), *zip(IDs, index)))
        me['_index'] = index
        me['_isheavy'] = np.array([i.Tb in (inf, None) for i in cmps])
        me['_islight'] = np.array([i.Tb in (0, -inf) for i in cmps], dtype=bool)
        nonfinite = (inf, -inf, None)
        me['_has_equilibrium'] = np.array([(bool(i.UNIFAC_Dortmund_groups)
                                            and i.Tb not in nonfinite)
                                           for i in cmps])
        me['_compounds'] = cmps
        me['_IDs'] = IDs
        me['_N'] = N
        me['_MW'] = np.array([i.MW for i in cmps])
        return self
    
    def __dir__(self):
        return [i for i in super().__dir__() if (i.isalnum() or (i and i[0]=='_'))]

    def get_synonyms(self, ID):
        dct = self._indexdct
        index = dct[ID]
        return [i for i, j in dct.items() if j==index] 

    def set_synonym(self, ID, synonym):
        compound = getattr(self, ID)
        dct = self.__dict__
        if synonym in dct and dct[synonym] is not compound:
            raise ValueError(f"synonym '{synonym}' already in use by {repr(dct[synonym])}")
        else:
            self._indexdct[synonym] = self._indexdct[ID]
            dct[synonym] = compound
            
    def kwarray(self, **data):
        """Return an array with entries that correspond to compound IDs.
        
        **Parameters**
        
            ****data:** ID-value pair.
            
        **Examples**
        
            >>> from biosteam import *
            >>> Stream.species = Species('Water', 'Ethanol')
            >>> Stream.species.kwarray(Water=2)
            array([2., 0.])
        
        """
        return self.array(data, [*data.values()])
    
    def array(self, IDs, data):
        """Return an array with entries that correspond to species IDs.
        
        **Parameters**
        
            **IDs:** [iterable] compound IDs.
            
            **data:** [array_like] Data corresponding to IDs.
            
        **Examples**
        
            >>> from biosteam import *
            >>> Stream.species = Species('Water', 'Ethanol')
            >>> Stream.species.array(['Water'], [2])
            array([2., 0.])
        
        """
        array = np.zeros(self._N)
        array[self.indices(IDs)] = data
        return array

    def index(self, ID):
        """Return index of specified compound.

        **Parameters**

             **ID:** [str] compound ID

        Example

        Index by ID:
        
        >>> from biosteam import *
        >>> Stream.species = Species(['Ethanol', 'Water'])
        >>> Stream.species.index('Water')
        1

        Indices by CAS number:
        
        >>> Stream.species.index('7732-18-5'):
        1

        """
        try: return self._indexdct[ID]
        except KeyError:
            if ID not in self._indexdct: raise UndefinedCompound(ID)

    def indices(self, IDs):
        """Return indices of specified species.

        **Parameters**

             **IDs:** [iterable] species IDs or CAS numbers.

        Example

        Indices by ID:
        
        >>> from biosteam import *
        >>> Stream.species = Species(['Ethanol', 'Water'])
        >>> Stream.species.indices(['Water', 'Ethanol'])
        [1, 0]

        Indices by CAS number:
        
        >>> Stream.species.indices(['7732-18-5', '64-17-5']):
        [1, 0]

        """
        try:
            dct = self._indexdct
            return [dct[i] for i in IDs]
        except KeyError:
            for i in IDs:
                if i not in dct: raise UndefinedCompound(i)       
    
    def __delattr__(self, ID):
        AttributeError(f"'{type(self).__name__} object is read-only")
    
    def __setattr__(self, ID, name):
        AttributeError(f"'{type(self).__name__} object is read-only")
    
    def __len__(self):
        return self._N
    
    def __contains__(self, compound):
        return compound in self._compounds
    
    def __iter__(self):
        yield from self._compounds
    
    @property
    def IDs(self):
        """[tuple] IDs of Species object."""
        return self._IDs

    ### Material Properties ###
    getprops = Species.getprops
    setprops = Species.setprops

    # General methods for getting ideal mixture properties
    def _props(self, ID, mol, T, P, phase):
        """Return component property list."""
        out = np.zeros(self._N)
        getattr_ = getattr
        cmps = self._compounds
        for i in self._index:
            if mol[i]: 
                c = cmps[i]
                c.P = P
                c.T = T
                c.phase = phase
                out[i] = getattr_(c, ID)
        return out

    def _propflows(self, ID, mol, T, P, phase):
        """Return array of component properties * kmol/hr."""
        return self._props(ID, mol, T, P, phase) * mol

    def _propflow(self, ID, mol, T, P, phase):
        """Return flow of component properties."""
        return (self._props(ID, mol, T, P, phase)*mol).sum()

    def _prop(self, ID, mol, T, P, phase):
        """Return molar weighted average property."""
        try:
            return (self._props(ID, mol, T, P, phase)*mol).sum()/mol.sum()
        except:
            # This usually fails due to a missing property.
            # Avoid this problem by ignoring the property.
            # TODO: Maybe include a warning.
            propflow = 0
            molnet = 0
            getattr_ = getattr
            cmps = self._compounds
            for i in self._index:
                if mol[i]:
                    c = cmps[i]
                    c.P = P
                    c.T = T
                    c.phase = phase
                    prop = getattr_(c, ID)
                    if prop:
                        molnet += mol[i]
                        propflow += prop*mol[i]
            return propflow/molnet

    def _mixedpropflows(self, ID, molarray, T, P):
        """Return array of component properties * kmol/hr."""
        pf = self._propflows
        return (pf(ID, molarray[0], T, P, 's')
              + pf(ID, molarray[1], T, P, 'l')
              + pf(ID, molarray[2], T, P, 'l')
              + pf(ID, molarray[3], T, P, 'g'))
    
    def _mixedpropflow(self, ID, molarray, T, P):
        """Return flow of component properties."""
        return self._mixedpropflows(ID, molarray, T, P).sum()
    
    def _mixedprop(self, ID, molarray, T, P):
        """Return molar weighted average property."""
        return self._mixedpropflows(ID, molarray, T, P).sum()/molarray.sum()  
    
    ### Equilibrium ###
    
    def _equilibrium_indices(self, nonzero):
        """Return indices of species in equilibrium."""
        condition = self._has_equilibrium & nonzero
        return [i for i in self._index if condition[i]]

    def _heavy_indices(self, nonzero):
        """Return indices of heavy species not in equilibrium."""
        condition = self._isheavy & nonzero
        return [i for i in self._index if condition[i]]

    def _light_indices(self, nonzero):
        """Return indices of light species not in equilibrium."""
        condition = self._islight & nonzero
        return [i for i in self._index if condition[i]]
    
    __repr__ = Species.__repr__

    



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 22:14:01 2018

@author: yoelr
"""
from ._solids_separator import SolidsSeparator
from .decorators import cost

@cost('Flow rate', units='kg/hr', cost=1.5e6, CE=541.7,
      n=0.6, S=335e3, kW=2010, BM=2.3)
class CrushingMill(SolidsSeparator):
    """Create CrushingMill object.
    
    **Parameters**
    
        **moisture_content:** Fraction of water in Baggasse
    
    **ins**
    
        [0] Shredded sugar cane
        
        [1] Recycle water
    
    **outs**
    
        [0] Bagasse
        
        [1] Juice
    
    """
    



# -*- coding: utf-8 -*-
"""
Created on Mon Apr 29 18:28:55 2019

@author: yoelr
"""
from . import Facility
from ..decorators import cost
# from copy import copy

__all__ = ('CoolingTower',) #'CoolingTowerWithPowerDemand')

@cost('Flow rate', 'Cooling water pump',
      S=557183, kW=1021, cost=283671, CE=551, n=0.8, BM=3.1)
@cost('Flow rate', 'Cooling tower',
      S=557183, kW=1598, cost=1375e3, CE=551, n=0.7, BM=1.5)
class CoolingTower(Facility):
    """Create a cooling tower that is cost based on flow rate of cooling water."""
    _units = {'Flow rate': 'kmol/hr'}
    _N_heat_utilities = 1
    _N_outs = _N_ins = 0
    def __init__(self, ID=''):
        Facility.__init__(self, ID, None, None)
        self.cooling_water_utilities = set()
        
    def _design(self):
        cwu = self.cooling_water_utilities
        if not cwu:
            for u in self.system.units:
                if u is self: continue
                for hu in u._heat_utilities:
                    if hu.ID == 'Cooling water':
                        cwu.add(hu)
        #: Cooling water flow rate (kmol/hr)
        self._Design['Flow rate'] = self.cooling_water = q = sum([i.flow for i in cwu])
        hu = self._heat_utilities[0]
        cw = hu.cooling_agents['Cooling water']
        hu.ID = 'Cooling water'
        hu.cost = -q*cw.price_kmol
        

   
# class CoolingTowerWithPowerDemand(CoolingTower):
#     _has_power_utility = True
#     _N_heat_utilities = 1
#     cost_options = copy(CoolingTower.cost_items)
#     cost_options['Cooling tower'].kW = 0.1
#     def _cost(self):
#         super()._cost()
#         q = self._molar_flow # kmol/hr
#         hu = self._heat_utilities[0]
#         cw = hu.cooling_agents['Cooling water']
#         hu.ID = 'Cooling water'
#         hu.flow = -q
#         hu.cost = -q*cw.price_kmol


# -*- coding: utf-8 -*-
"""
Created on Wed Mar 20 18:40:05 2019

@author: yoelr
"""
from .._utils import bounded_wegstein, wegstein, aitken, \
                     bounded_aitken, IQ_interpolation
import numpy as np

__all__ = ('VLE', 'V_2N', 'V_3N', 'V_error')

def V_error(V, zs, Ks):
    """Vapor fraction error."""
    return (zs*(Ks-1.)/(1.+V*(Ks-1.))).sum()

def V_2N(zs, Ks):
    """Solution for 2 component flash vessel."""
    z1, z2 = zs
    K1, K2 = Ks
    return (-K1*z1 - K2*z2 + z1 + z2)/(K1*K2*z1 + K1*K2 *
                                       z2 - K1*z1 - K1*z2
                                       - K2*z1 - K2*z2 + z1 + z2)
    
def V_3N(zs, Ks):
    """Solution for 3 component flash vessel."""
    z1, z2, z3 = zs
    K1, K2, K3 = Ks
    return (-K1*K2*z1/2 - K1*K2*z2/2 - K1*K3*z1/2 - K1*K3*z3/2 + K1*z1 + K1*z2/2 + K1*z3/2 - K2*K3*z2/2 - K2*K3*z3/2 + K2*z1/2 + K2*z2 + K2*z3/2 + K3*z1/2 + K3*z2/2 + K3*z3 - z1 - z2 - z3 - (K1**2*K2**2*z1**2 + 2*K1**2*K2**2*z1*z2 + K1**2*K2**2*z2**2 - 2*K1**2*K2*K3*z1**2 - 2*K1**2*K2*K3*z1*z2 - 2*K1**2*K2*K3*z1*z3 + 2*K1**2*K2*K3*z2*z3 - 2*K1**2*K2*z1*z2 + 2*K1**2*K2*z1*z3 - 2*K1**2*K2*z2**2 - 2*K1**2*K2*z2*z3 + K1**2*K3**2*z1**2 + 2*K1**2*K3**2*z1*z3 + K1**2*K3**2*z3**2 + 2*K1**2*K3*z1*z2 - 2*K1**2*K3*z1*z3 - 2*K1**2*K3*z2*z3 - 2*K1**2*K3*z3**2 + K1**2*z2**2 + 2*K1**2*z2*z3 + K1**2*z3**2 - 2*K1*K2**2*K3*z1*z2 + 2*K1*K2**2*K3*z1*z3 - 2*K1*K2**2*K3*z2**2 - 2*K1*K2**2*K3*z2*z3 - 2*K1*K2**2*z1**2 - 2*K1*K2**2*z1*z2 - 2*K1*K2**2*z1*z3 + 2*K1*K2**2*z2*z3 + 2*K1*K2*K3**2*z1*z2 - 2*K1*K2*K3**2*z1*z3 - 2*K1*K2*K3**2*z2*z3 - 2*K1*K2*K3**2*z3**2 + 4*K1*K2*K3*z1**2 + 4*K1*K2*K3*z1*z2 + 4*K1*K2*K3*z1*z3 + 4*K1*K2*K3*z2**2 + 4*K1*K2*K3*z2*z3 + 4*K1*K2*K3*z3**2 + 2*K1*K2*z1*z2 - 2*K1*K2*z1*z3 - 2*K1*K2*z2*z3 - 2*K1*K2*z3**2 - 2*K1*K3**2*z1**2 - 2*K1*K3**2*z1*z2 - 2*K1*K3**2*z1*z3 + 2*K1*K3**2*z2*z3 - 2*K1*K3*z1*z2 + 2*K1*K3*z1*z3 - 2*K1*K3*z2**2 - 2*K1*K3*z2*z3 + K2**2*K3**2*z2**2 + 2*K2**2*K3**2*z2*z3 + K2**2*K3**2*z3**2 + 2*K2**2*K3*z1*z2 - 2*K2**2*K3*z1*z3 - 2*K2**2*K3*z2*z3 - 2*K2**2*K3*z3**2 + K2**2*z1**2 + 2*K2**2*z1*z3 + K2**2*z3**2 - 2*K2*K3**2*z1*z2 + 2*K2*K3**2*z1*z3 - 2*K2*K3**2*z2**2 - 2*K2*K3**2*z2*z3 - 2*K2*K3*z1**2 - 2*K2*K3*z1*z2 - 2*K2*K3*z1*z3 + 2*K2*K3*z2*z3 + K3**2*z1**2 + 2*K3**2*z1*z2 + K3**2*z2**2)**0.5/2)/(K1*K2*K3*z1 + K1*K2*K3*z2 + K1*K2*K3*z3 - K1*K2*z1 - K1*K2*z2 - K1*K2*z3 - K1*K3*z1 - K1*K3*z2 - K1*K3*z3 + K1*z1 + K1*z2 + K1*z3 - K2*K3*z1 - K2*K3*z2 - K2*K3*z3 + K2*z1 + K2*z2 + K2*z3 + K3*z1 + K3*z2 + K3*z3 - z1 - z2 - z3)

class VLE:
    """Create a VLE object for solving VLE."""
    __slots__ = ('T', 'P', 'Q', 'V', '_stream', '_gamma',
                 '_dew_point', '_bubble_point', '_v', '_liquid_mol',
                 '_vapor_mol', '_index', '_massnet', '_compound',
                 '_update_V', '_mol', '_molnet', '_N', '_solve_V',
                 '_zs', '_Ks', '_Psat_gama', '_Psat_P')
    
    solver = staticmethod(IQ_interpolation)
    itersolver = staticmethod(aitken)
    T_tol = 0.00001
    P_tol = 0.1
    Q_tol = 0.1
    V_tol = 0.00001
    
    def __init__(self, stream):
        self.T = self.P = self.Q = self.V = 0
        self._stream = stream
        self._dew_point = stream._dew_point
        self._bubble_point = stream._bubble_point
        self._gamma = stream._gamma
        self._liquid_mol = stream.liquid_mol
        self._vapor_mol = stream.vapor_mol
    
    def __call__(self, species_IDs=None, LNK=None, HNK=None,
                 P=None, Q=None, T=None, V=None, x=None, y=None):
        """Perform vapor-liquid equilibrium.

        **Parameters**
        
            **Specify two:**
                * **P:** [float] Operating pressure (Pa)
                * **Q:** [float] Energy input (kJ/hr)
                * **T:** [float] Operating temperature (K)
                * **V:** [float] Molar vapor fraction
                * **x:** [numpy array] Molar composition of liquid (for binary mixture)
                * **y:** [numpy array] Molar composition of vapor (for binary mixture)

            **species_IDs:** [tuple] IDs of species in equilibrium.
                 
            **LNK:** tuple[str] Light non-keys that remain as a vapor.
    
            **HNK:** tuple[str] Heavy non-keys that remain as a liquid.

        .. Note:
           LNK and HNK are not taken into account for equilibrium.

        """
        ### Decide what kind of equilibrium to run ###
        T_spec = T is not None
        P_spec = P is not None
        V_spec = V is not None
        Q_spec = Q is not None
        x_spec = x is not None
        y_spec = y is not None
        N_specs = (P_spec + Q_spec + T_spec + V_spec + x_spec + y_spec)
        assert N_specs == 2, ("must pass two and only two of the following specifications: "
                              "T, P, V, Q, x, y")
        self.setup(species_IDs, LNK, HNK)
        if T_spec:
            if P_spec:
                return self.TP(T, P)
            elif V_spec:
                return self.TV(T, V)
            elif Q_spec:
                return self.TQ(T, Q)
            elif x_spec:
                return self.Tx(T, np.asarray(x))
            else: # y_spec
                return self.Ty(T, np.asarray(y))
        elif P_spec:
            if V_spec:
                return self.PV(P, V)
            elif Q_spec:
                return self.PQ(P, Q)
            elif x_spec:
                return self.Px(P, np.asarray(x))
            else: # y_spec
                return self.Py(P, np.asarray(y))
        elif Q_spec:
            if y_spec:
                raise NotImplementedError('specification Q and y not implemented')
            elif x_spec:
                raise NotImplementedError('specification Q and x not implemented')
            else: # V_spec
                raise NotImplementedError('specification V and Q not implemented yet')
        else: # x_spec and y_spec
            raise ValueError("can only pass either 'x' or 'y' arguments, not both")
    
    def setup(self, species_IDs, LNK, HNK):
        liquid_mol = self._liquid_mol 
        vapor_mol = self._vapor_mol

        # Get flow rates
        mol = liquid_mol + vapor_mol
        notzero = mol > 0

        # Reused attributes
        species = self._stream._species
        speciesdct = species.__dict__
        indexdct = species._indexdct

        # Set up indices for both equilibrium and non-equilibrium species
        if species_IDs:
            index = [indexdct[specie] for specie in species_IDs]
            self._gamma.species = [speciesdct[ID] for ID in species_IDs]
        else:
            index = species._equilibrium_indices(notzero)
            cmps = species._compounds
            self._gamma.species = [cmps[i] for i in index]
        if LNK:
            LNK_index = [indexdct[i] for i in LNK]
        else:
            LNK_index = species._light_indices(notzero)
        if HNK:
            HNK_index = [indexdct[i] for i in HNK]
        else:
            HNK_index = species._heavy_indices(notzero)
        self._mol = mol[index]
        self._index = index

        # Set light and heavy keys
        vapor_mol[HNK_index] = 0
        vapor_mol[LNK_index] = mol[LNK_index]
        liquid_mol[LNK_index] = 0
        liquid_mol[HNK_index] = mol[HNK_index]
        
        self._N = N = len(index)
        if N == 1:
            self._compound = self._gamma._species[0]
            return 
        elif N == 2:
            self._solve_V = self._solve_V_2
        elif N == 3:
            self._solve_V = self._solve_V_3
        else:
            self._solve_V = self._solve_V_N
        self._massnet = self._stream.massnet
        
        # Get overall composition
        self._molnet = molnet = self._mol.sum()
        assert molnet != 0, 'empty stream cannot perform equilibrium'
        self._zs = self._mol/molnet

    ### Single component equilibrium case ###
        
    def _TP_compound(self, T, P):
        # Either liquid or gas
        if P < self._compound.VaporPressure(T):
            self._liquid_mol[self.index] = 0
            self._vapor_mol[self.index] = self._mol
        else:
            self._liquid_mol[self._index] = self._mol
            self._vapor_mol[self._index] = 0
    
    def _TV_compound(self, T, V):
        # Set vapor fraction
        self._stream.P = self._compound.VaporPressure(T)
        self._vapor_mol[self._index] = self._mol*V
        self._liquid_mol[self._index] = self._mol - self._vapor_mol[self._index]
        
    def _PV_compound(self, P, V):
        # Set vapor fraction
        self.T = self._compound.Tsat(P)
        self._vapor_mol[self._index] = self._mol*V
        self._liquid_mol[self._index] = self._mol - self._vapor_mol[self._index]
        
    def _PQ_compound(self, P, Q): 
        mol = self._mol
        index = self._index
        stream = self._stream
        vapor_mol = self._vapor_mol
        liquid_mol = self._liquid_mol
        Hin = Q + stream.H
        
        # Set temperature in equilibrium
        stream.T = self._compound.Tsat(P)
        
        # Check if super heated vapor
        vapor_mol[index] = mol
        liquid_mol[index] = 0
        H_dew = stream.H
        if Hin >= H_dew:
            stream.H = Hin
            return

        # Check if subcooled liquid
        vapor_mol[index] = 0
        liquid_mol[index] = mol
        H_bubble = stream.H
        if Hin <= H_bubble:
            stream.H = Hin
            return
        
        # Adjust vapor fraction accordingly
        V = (Hin - H_bubble)/(H_dew - H_bubble)
        vapor_mol[index] = mol*V
        liquid_mol[index] = mol - vapor_mol[index]
        
    def _TQ_compound(self, T, Q):
        index = self._index
        mol = self._mol
        stream = self._stream
        vapor_mol = self._vapor_mol
        liquid_mol = self._liquid_mol
        Hin = Q + stream.H
        
        # Set Pressure in equilibrium
        stream.P = self._compound.VaporPressure(T)
        
        # Check if super heated vapor
        vapor_mol[index] = mol
        liquid_mol[index] = 0
        H_dew = stream.H
        if Hin >= H_dew:
            stream.H = Hin
            return

        # Check if subcooled liquid
        vapor_mol[index] = 0
        liquid_mol[index] = mol
        H_bubble = stream.H
        if Hin <= H_bubble:
            stream.H = Hin
            return
        
        # Adjust vapor fraction accordingly
        V = (Hin - H_bubble)/(H_dew - H_bubble)
        vapor_mol[index] = mol*V
        liquid_mol[index] = mol - vapor_mol[index]
        
    def _lever_rule(self, x, y):
        split_frac = (self._zs[0]-x[0])/(y[0]-x[0])
        assert -0.00001 < split_frac < 1.00001, 'desired composition is infeasible'
        if split_frac > 1:
            split_frac = 1
        elif split_frac < 0:
            split_frac = 0
        self._vapor_mol[self._index] = self._molnet * split_frac * y
        self._liquid_mol[self._index] = self._mol - self._vapor_mol[self._index]
    
    def Tx(self, T, x):
        assert self._N == 2, 'number of species in equilibrium must be 2 to specify x'
        self._stream.P, y = self._bubble_point.solve_Px(x, T)
        self._lever_rule(x, y)
    
    def Px(self, P, x):
        assert self._N == 2, 'number of species in equilibrium must be 2 to specify x'
        self._stream.T, y = self._bubble_point.solve_Ty(x, P) 
        self._lever_rule(x, y)
        
    def Ty(self, T, y):
        assert self._N == 2, 'number of species in equilibrium must be 2 to specify y'
        self._stream.P, x = self._dew_point.solve_Px(y, T)
        self._lever_rule(x, y)
    
    def Py(self, P, y):
        assert self._N == 2, 'number of species in equilibrium must be 2 to specify y'
        self._stream.T, x = self._dew_point.solve_Ty(y, P) 
        self._lever_rule(x, y)
        
    def TP(self, T, P):
        if self._N == 1: self._TP_compound(T, P)
        stream = self._stream
        self.T = stream.T = T
        self.P = self._stream.P = P
        # Setup bounderies
        P_dew, x_dew = self._dew_point.solve_Px(self._zs, T)
        P_bubble, y_bubble = self._bubble_point.solve_Py(self._zs, T)
        
        # Check if there is equilibrium
        if T >= P_dew:
            self._vapor_mol[self._index] = self._mol
            self._liquid_mol[self._index] = 0
        elif T <= P_bubble:
            self._vapor_mol[self._index] = 0
            self._liquid_mol[self._index] = self._mol
        else:
            # Guess composition in the vapor is a
            # weighted average of bubble/dew points
            self.V = V = self.V or (T - P_dew)/(P_bubble - P_dew)
            y = V*self._zs + (1-V)*y_bubble
            
            # Guess vapor flow rates
            self._v = y * V * self._mol

            # Solve
            self._vapor_mol[self._index] = self._solve_v(T, P)
            self._liquid_mol[self._index] = self._mol - self._v
            self.Q = stream.H/self._massnet
        
    def TV(self, T, V):
        if self._N == 1: self._TV_compound(T, V)
        stream = self._stream
        self.T = stream.T = T
        P_dew, x_dew = self._dew_point.solve_Px(self._zs, T)
        P_bubble, y_bubble = self._bubble_point.solve_Py(self._zs, T)
        if V == 1:
            self._stream.P = P_dew
            self._vapor_mol[self._index] = self._mol
            self._liquid_mol[self._index] = 0
        elif V == 0:
            self._stream.P = P_bubble
            self._vapor_mol[self._index] = 0
            self._liquid_mol[self._index] = self._mol
        else:
            self.V = V
            self._v = (V*self._zs + (1-V)*y_bubble)*V*self._molnet
            self.P = stream.P = self.solver(self._V_at_P,
                                            P_bubble, P_dew, 0, 1,
                                            self.P, self.V,
                                            self.P_tol, self.V_tol)
            self._vapor_mol[self._index] = self._v
            self._liquid_mol[self._index] = self._mol - self._v
            self.Q = stream.H/self._massnet

    def TQ(self, T, Q):
        if self._N == 1: self._TQ_compound(T, Q)
        stream = self._stream
        stream.T = T
        
        # Setup bounderies
        P_dew, x_dew = self._dew_point.solve_Px(self._zs, T)
        P_bubble, y_bubble = self._bubble_point.solve_Py(self._zs, T)
        index = self._index
        mol = self._mol
        vapor_mol = self._vapor_mol
        liquid_mol = self._liquid_mol
        Hin = Q + stream.H
        
        # Check if super heated vapor
        vapor_mol[index] = mol
        liquid_mol[index] = 0
        stream.P = P_dew
        H_dew = stream.H
        dH_dew = (Hin - H_dew)
        if dH_dew >= 0:
            stream.H = Hin
            return

        # Check if subcooled liquid
        vapor_mol[index] = 0
        liquid_mol[index] = mol
        stream.P = P_bubble
        H_bubble = stream.H
        dH_bubble = (Hin - H_bubble)
        if dH_bubble <= 0:
            stream.H = Hin
            return

        # Guess overall vapor fraction, and vapor flow rates
        V = self.V or dH_bubble/(H_dew - H_bubble)
        # Guess composition in the vapor is a weighted average of boiling points
        self._v = V*self._zs + (1-V)*y_bubble*V*self._molnet
        massnet = self._massnet
        self.Q = Hin/massnet
        self.P = stream.P = self.solver(self._Q_at_P,
                                        P_bubble, P_dew,
                                        H_bubble/massnet, H_dew/massnet,
                                        self.P, self.Q,
                                        self.P_tol, self.Q_tol) 
        self.T = T
    
    def PV(self, P, V):
        if self._N == 1: self._PV_compound(P, V)
        # Setup bounderies
        T_dew, x_dew = self._dew_point.solve_Tx(self._zs, P)
        T_bubble, y_bubble = self._bubble_point.solve_Ty(self._zs, P)
        
        stream = self._stream
        stream.P = P
        index = self._index
        mol = self._mol
        vapor_mol = self._vapor_mol
        liquid_mol = self._liquid_mol
        
        if V == 1:
            stream.T = T_dew
            vapor_mol[index] = mol
            liquid_mol[index] = 0
            return
        elif V == 0:
            stream.T = T_bubble
            vapor_mol[index] = 0
            liquid_mol[index] = mol
            return
        else:
            self._v = (V*self._zs + (1-V)*y_bubble) * V * self._molnet
            self.V = V
            self.T = stream.T = self.solver(self._V_at_T,
                                            T_bubble, T_dew, 0, 1,
                                            self.T, V,
                                            self.T_tol, self.V_tol)
            vapor_mol[index] = self._v
            liquid_mol[index] = mol - self._v
            self.P = P
            self.Q = stream.H/self._massnet
    
    def PQ(self, P, Q):
        if self._N == 1: self._PQ_compound(P, Q)
        # Setup bounderies
        T_dew, x_dew = self._dew_point.solve_Tx(self._zs, P)
        T_bubble, y_bubble = self._bubble_point.solve_Ty(self._zs, P)
        
        stream = self._stream
        stream.P = P
        index = self._index
        mol = self._mol
        vapor_mol = self._vapor_mol
        liquid_mol = self._liquid_mol
        Hin = Q + stream.H
        
        # Check if super heated vapor
        vapor_mol[index] = mol
        liquid_mol[index] = 0
        stream.T = T_dew
        H_dew = stream.H
        dH_dew = Hin - H_dew
        if dH_dew >= 0:
            stream.H = Hin
            return

        # Check if subcooled liquid
        vapor_mol[index] = 0
        liquid_mol[index] = mol
        stream.T = T_bubble
        H_bubble = stream.H
        dH_bubble = Hin - H_bubble
        if dH_bubble <= 0:
            stream.H = Hin
            return
        
        # Guess T, overall vapor fraction, and vapor flow rates
        self.V = V = self.V or dH_bubble/(H_dew - H_bubble)
        self._v = V*self._zs + (1-V)*y_bubble*V*self._molnet
        
        massnet = self._massnet
        self.Q = Hin/massnet
        self.T = stream.T = self.solver(self._Q_at_T,
                                        T_bubble, T_dew, 
                                        H_bubble/massnet, H_dew/massnet,
                                        self.T, self.Q,
                                        self.T_tol, self.Q_tol)
        self.P = P
    
    def _Q_at_T(self, T):
        self._stream.T = T
        self._vapor_mol[self._index] = self._solve_v(T, self._stream.P)
        self._liquid_mol[self._index] = self._mol - self._v
        return self._stream.H/self._massnet
    
    def _Q_at_P(self, P):
        self._stream.P = P
        self._vapor_mol[self._index] = self._solve_v(self._stream.T, P)
        self._liquid_mol[self._index] = self._mol - self._v
        return self._stream.H/self._massnet
    
    def _V_at_P(self, P):
        return self._solve_v(self._stream.T, P).sum()/self._molnet
    
    def _V_at_T(self, T):
        return self._solve_v(T, self._stream.P).sum()/self._molnet
    
    def _x_iter(self, x):
        self._Ks = self._Psat_P * self._gamma(x/x.sum(), self._stream.T)
        return self._zs/(1. + self._solve_V()*(self._Ks-1.))
    
    def _solve_v(self, T, P):
        """Solve for vapor mol"""
        gamma = self._gamma
        self._Psat_P = np.array([s.VaporPressure(T) for s in gamma.species])/P
        l = self._mol - self._v
        x = self.itersolver(self._x_iter, l/l.sum(), 1e-4)
        self._v = self._molnet*self.V*x/x.sum()*self._Ks            
        return self._v

    def _V_error(self, V):
        """Vapor fraction error."""
        return (self._zs*(self._Ks-1.)/(1.+V*(self._Ks-1.))).sum()

    def _solve_V_N(self):
        """Update V for N components."""
        self.V = self.solver(self._V_error, 0, 1,
                             self._V_error(0), self._V_error(1),
                             self.V, 0, 1e-4, 1e-7)
        return self.V

    def _solve_V_2(self):
        """Update V for 2 components."""
        z1, z2 = self._zs
        K1, K2 = self._Ks
        self.V = (-K1*z1 - K2*z2 + z1 + z2)/(K1*K2*z1 + K1*K2 *
                                             z2 - K1*z1 - K1*z2
                                             - K2*z1 - K2*z2 + z1 + z2)
        return self.V
    
    def _solve_V_3(self):
        """Update V for 3 components."""
        z1, z2, z3 = self._zs
        K1, K2, K3 = self._Ks
        self.V = ((-K1*K2*z1/2 - K1*K2*z2/2 - K1*K3*z1/2 - K1*K3*z3/2 + K1*z1
                   + K1*z2/2 + K1*z3/2 - K2*K3*z2/2 - K2*K3*z3/2 + K2*z1/2
                   + K2*z2 + K2*z3/2 + K3*z1/2 + K3*z2/2 + K3*z3 - z1 - z2 - z3
                   - (K1**2*K2**2*z1**2 + 2*K1**2*K2**2*z1*z2 + K1**2*K2**2*z2**2
                      - 2*K1**2*K2*K3*z1**2 - 2*K1**2*K2*K3*z1*z2 - 2*K1**2*K2*K3*z1*z3
                      + 2*K1**2*K2*K3*z2*z3 - 2*K1**2*K2*z1*z2 + 2*K1**2*K2*z1*z3
                      - 2*K1**2*K2*z2**2 - 2*K1**2*K2*z2*z3 + K1**2*K3**2*z1**2
                      + 2*K1**2*K3**2*z1*z3 + K1**2*K3**2*z3**2 + 2*K1**2*K3*z1*z2
                      - 2*K1**2*K3*z1*z3 - 2*K1**2*K3*z2*z3 - 2*K1**2*K3*z3**2
                      + K1**2*z2**2 + 2*K1**2*z2*z3 + K1**2*z3**2 - 2*K1*K2**2*K3*z1*z2
                      + 2*K1*K2**2*K3*z1*z3 - 2*K1*K2**2*K3*z2**2 - 2*K1*K2**2*K3*z2*z3
                      - 2*K1*K2**2*z1**2 - 2*K1*K2**2*z1*z2 - 2*K1*K2**2*z1*z3
                      + 2*K1*K2**2*z2*z3 + 2*K1*K2*K3**2*z1*z2 - 2*K1*K2*K3**2*z1*z3
                      - 2*K1*K2*K3**2*z2*z3 - 2*K1*K2*K3**2*z3**2 + 4*K1*K2*K3*z1**2
                      + 4*K1*K2*K3*z1*z2 + 4*K1*K2*K3*z1*z3 + 4*K1*K2*K3*z2**2
                      + 4*K1*K2*K3*z2*z3 + 4*K1*K2*K3*z3**2 + 2*K1*K2*z1*z2
                      - 2*K1*K2*z1*z3 - 2*K1*K2*z2*z3 - 2*K1*K2*z3**2 - 2*K1*K3**2*z1**2
                      - 2*K1*K3**2*z1*z2 - 2*K1*K3**2*z1*z3 + 2*K1*K3**2*z2*z3
                      - 2*K1*K3*z1*z2 + 2*K1*K3*z1*z3 - 2*K1*K3*z2**2 - 2*K1*K3*z2*z3
                      + K2**2*K3**2*z2**2 + 2*K2**2*K3**2*z2*z3 + K2**2*K3**2*z3**2
                      + 2*K2**2*K3*z1*z2 - 2*K2**2*K3*z1*z3 - 2*K2**2*K3*z2*z3
                      - 2*K2**2*K3*z3**2 + K2**2*z1**2 + 2*K2**2*z1*z3
                      + K2**2*z3**2 - 2*K2*K3**2*z1*z2 + 2*K2*K3**2*z1*z3
                      - 2*K2*K3**2*z2**2 - 2*K2*K3**2*z2*z3 - 2*K2*K3*z1**2
                      - 2*K2*K3*z1*z2 - 2*K2*K3*z1*z3 + 2*K2*K3*z2*z3 + K3**2*z1**2
                      + 2*K3**2*z1*z2 + K3**2*z2**2)**0.5/2)
                   / (K1*K2*K3*z1 + K1*K2*K3*z2 + K1*K2*K3*z3 - K1*K2*z1
                      - K1*K2*z2 - K1*K2*z3 - K1*K3*z1 - K1*K3*z2 - K1*K3*z3
                      + K1*z1 + K1*z2 + K1*z3 - K2*K3*z1 - K2*K3*z2 - K2*K3*z3
                      + K2*z1 + K2*z2 + K2*z3 + K3*z1 + K3*z2 + K3*z3 - z1 - z2 - z3))
        return self.V

    def __repr__(self):
        return f"{repr(self._stream)}.VLE"



# -*- coding: utf-8 -*-
"""
Created on Thu Mar  7 23:17:38 2019

@author: yoelr
"""
import numpy as np
import pandas as pd
from ._state import State

__all__ = ('Model',)

# %% Functions

paramindex = lambda params: pd.MultiIndex.from_arrays(
                                ([f.element_name for f in params], 
                                 [f.name for f in params]),
                                names=('Element', 'Parameter'))

    
# %% Grid of simulation blocks

class Model(State):
    """Create a Model object that allows for evaluation over a sample space.
    
    **Parameters**
    
        **system:** [System] Should reflect the model state.
    
        **skip:** [bool] If True, skip simulation for repeated states
    
        ****metrics:** [str: function] ID-metric pairs.
    
    **Examples**

         :doc:`Advanced simulation`
    
    """
    __slots__ = ('_table',   # [DataFrame] All arguments and results
                 '_metrics', # dict[ID: function] Functions should return evaluated metric
                 '_index',   # list[int] Order of sample evaluation for performance
                 '_samples', # [array] Argument sample space
                 '_setters', # list[function] Cached parameter setters
                 '_repeats') # [bool] True if parameter states repeat
    
    def __init__(self, system, skip=False, **metrics):
        super().__init__(system, skip)
        self._metrics = metrics
        self._repeats = self._setters = self._samples = self._table = None
    
    def _erase(self):
        """Erase cached data."""
        self._setters = self._model = self._table = self._samples = None
    
    @property
    def metrics(self):
        """dict[str: function] ID-metric pairs."""
        return self._metrics
    @metrics.setter
    def metrics(self, metrics):
        table = self._table
        if table is not None:
            for i in self._metrics:
                if i in table: del table[i]
        self._metrics = metrics
    
    @property
    def table(self):
        """[DataFrame] Table of the sample space with results in the final column."""
        return self._table
    
    def load_samples(self, samples, has_repeats=None):
        """Load samples for evaluation"""
        if not self._model: self._loadmodel()
        params = self._params
        paramlen = len(params)
        if not isinstance(samples, np.ndarray):
            raise TypeError(f'samples must be an ndarray, not a {type(samples).__name__} object')
        elif samples.ndim != 2:
            raise ValueError('samples must be 2 dimensional')
        elif samples.shape[1] != paramlen:
            raise ValueError(f'number of parameters in samples ({samples.shape[1]}) must be equal to the number of parameters ({len(params)})')
        key = lambda x: samples[x][i]
        index = list(range(len(samples)))
        for i in range(paramlen-1,  -1, -1):
            if not params[i].system: break
            index.sort(key=key)
        self._index = index
        self._table = pd.DataFrame(samples, columns=paramindex(params))
        self._samples = samples
        if has_repeats is None:
            try:
                j = samples[0]
                for i in samples[1:]:
                    if any(i==j):
                        self._repeats = True
                        break
            except: pass
        else:
            self._repeats = has_repeats
        
    def evaluate(self, default=None):
        """Evaluate metric over the argument sample space and save values to `table`."""
        # Setup before simulation
        funcs = tuple(self._metrics.values())
        values = []
        add = values.append
        samples = self._samples
        if samples is None: raise ValueError('must load samples or distribution before evaluating')
        if self._repeats:
            model = self._model
            for i in self._index: 
                try: 
                    model(samples[i])
                    add([i() for i in funcs])
                except Exception as err:
                    if default: add(default)
                    else: raise err            
        else:
            if self._setters:
                setters = self._setters
            else:
                self._setters = setters = [p.setter for p in self._params]
            add = values.append
            sim = self._system.simulate
            for i in self._index:
                for f, s in zip(setters, samples[i]): f(s)
                sim()
                add([i() for i in funcs])
        for k, v in zip(self._metrics.keys(), zip(*values)): self.table[k] = v
    
    def __call__(self, sample):
        """Return dictionary of metric values."""
        super().__call__(sample)
        return {i:j() for i,j in self._metrics.items()}
    
    def _repr(self):
        clsname = type(self).__name__
        newline = "\n" + " "*(len(clsname)+2)
        return f'{clsname}: {newline.join(self._metrics.keys())}'
        
    def __repr__(self):
        return f'<{type(self).__name__}: {", ".join(self._metrics.keys())}>'
    
        


# -*- coding: utf-8 -*-
"""
Created on Thu Aug 23 15:53:14 2018

@author: yoelr
"""
import numpy as np
from fluids.pump import nema_sizes_hp
from .designtools._vacuum import _calc_MotorEfficiency, _calc_BreakEfficiency
from ._static import Static
import biosteam as bst

ln = np.log
exp = np.exp

# %% Data

max_hp = nema_sizes_hp[-1]

# Material factors
F_Mdict = {'Cast iron':       1,
           'Ductile iron':    1.15,
           'Cast steel':      1.35,
           'Bronze':          1.9,
           'Stainless steel': 2,
           'Hastelloy C':     2.95,
           'Monel':           3.3,
           'Nickel':          3.5,
           'Titanium':        9.7}

# Gear factor 
F_Tgear = {'OpenDripProof':           1,
           'EnclosedFanCooled':       1.4,
           'ExplosionProofEnclosure': 1.8}

# Centrifugal factor
F_Tcentrifugal = {'VSC3600':   1,
                  'VSC1800':   1.5,
                  'HSC3600':   1.7,
                  'HSC1800':   2,
                  '2HSC3600':  2.7,
                  '2+HSC3600': 8.9}

# %% Pump selection

# Pump types
Types = ('CentrifugalSingle', 'CentrifugalDouble', 'Gear')


def calc_NPSH(P_suction, P_vapor, rho_liq):
    """Return NPSH in ft given suction and vapor pressure in Pa and density in kg/m^3."""
    # Note: NPSH = (P_suction - P_vapor)/(rho_liq*gravity)
    # Taking into account units, NPSH will be equal to return value
    return 0.334438*(P_suction - P_vapor)/rho_liq
    

# %% Classes

# TODO: Fix pump selection to include NPSH available and required.
class Pump(Static):
    """Create a pump that sets the pressure of the 0th output stream.

    **Parameters**

        **P:** [float] Pressure of output stream (Pa). If None, cost pump as increase of pressure to P_startup.
    
    **ins**
    
        [0] Input stream
        
    **outs**
    
        [1] Output stream
    
    **Examples**
    
        :doc:`Pump Example`
    
    **References**
    
        [0] Seider, Warren D., et al. (2017). "Cost Accounting and Capital Cost Estimation". In Product and Process Design Principles: Synthesis, Analysis, and Evaluation (pp. 450-455). New York: Wiley.
    
    """
    _units = {'Ideal power': 'hp',
              'Power': 'hp',
              'Head': 'ft',
              'NPSH': 'ft',
              'Flow rate': 'gpm'}
    _has_power_utility = True
    BM = 3.3
    
    # Pump type
    _Type = 'Default'
    
    # Material factor
    _F_Mstr = 'Cast iron'
    _F_M = 1
    
    # Gear factor 
    _F_Tstr = 'VSC3600'
    _F_T = 1
    
    #: Pressure for costing purposes (Pa).
    P_startup = 405300 
    
    #: [bool] Ignore minimum NPSH requirement if True.
    ignore_NPSH = True

    @property
    def Type(self):
        """Pump type"""
        return self._Type
    @Type.setter
    def Type(self, Type):
        if Type not in Types:
            Types_str = str(Types)[1:-1]
            raise ValueError('Type must be one of the following: ' + Types_str)
        self._Type = Type
    
    @property
    def material(self):
        """Pump material"""
        return self._material
    @material.setter    
    def material(self, material):
        try:
            self._F_M = F_Mdict[material]
        except KeyError:
            dummy = str(F_Mdict.keys())[11:-2]
            raise ValueError(f"material must be one of the following: {dummy}")
        self._F_Mstr = material  
        
    def __init__(self, ID='', ins=None, outs=(), P=None):
        super().__init__(ID, ins, outs)
        self.P = P
        
    def _run(self):
        out = self._outs[0]
        feed = self._ins[0]
        out.P = self.P or feed.P
        out.T = feed.T
        out._phase = feed._phase
    
    def _design(self):
        Design = self._Design
        si = self.ins[0]
        so = self.outs[0]
        Pi = si.P
        Po = so.P
        Qi = si.volnet
        mass = si.massnet
        nu = si.nu
        dP = Po - Pi
        if dP < 1: dP = self.P_startup - Pi
        Design['Ideal power'] = power_ideal = Qi*dP*3.725e-7 # hp
        Design['Flow rate'] = q = Qi*4.403 # gpm
        if power_ideal <= max_hp:
            Design['Efficiency'] = e = self._calc_Efficiency(q, power_ideal)
            Design['Actual power'] = power = power_ideal/e
            Design['Pump power'] = self._nearest_PumpPower(power)
            Design['N'] = 1
            Design['Head'] = head = power/mass*897806 # ft
            # Note that:
            # head = power / (mass * gravity)
            # head [ft] = power[hp]/mass[kg/hr]/9.81[m/s^2] * conversion_factor
            # and 897806 = (conversion_factor/9.81)
        else:
            raise NotImplementedError('more than 1 pump required, but not yet implemented')
        
        if self.ignore_NPSH:
            NPSH_satisfied = True
        else:
            Design['NPSH'] = NPSH = calc_NPSH(Pi, si.P_vapor, si.rho)
            NPSH_satisfied = NPSH > 1.52
        
        # Get type
        Type = self.Type
        if Type == 'Default':
            if (0.00278 < q < 5000
                and 15.24 < head < 3200
                and nu < 0.00002
                and NPSH_satisfied):
                Type = 'Centrifugal'
            elif (q < 1500
                  and head < 914.4
                  and 0.00001 < nu < 0.252):
                Type = 'Gear'
            elif (head < 20000
                  and q < 500
                  and power < 200
                  and nu < 0.01):
                Type = 'MeteringPlunger'
            else:
                NPSH = calc_NPSH(Pi, si.P_vapor, si.rho)
                raise NotImplementedError(f'no pump type available at current power ({power:.3g} hp), flow rate ({q:.3g} gpm), and head ({head:.3g} ft), kinematic viscosity ({nu:.3g} m2/s), and NPSH ({NPSH:.3g} ft)')
                
        Design['Type'] = Type
        self._power_utility(power/1.341) # Set power in kW
    
    def _cost(self):
        # Parameters
        Design = self._Design
        Cost = self._Cost
        Type = Design['Type']
        q = Design['Flow rate']
        h = Design['Head']
        p = Design['Pump power']
        F_M = self._F_M
        I = bst.CE/567
        lnp = ln(p)
        
        # TODO: Add cost equation for small pumps
        # Head and flow rate is too small, so make conservative estimate on cost
        if q < 50: q = 50
        if h < 50: h = 50
        
        # Cost pump
        if 'Centrifugal' in Type:
            # Find pump factor
            F_Tdict = F_Tcentrifugal
            F_T = 1 # Assumption
            if p < 75 and 50 <= q <= 900 and 50 <= h <= 400:
                F_T = F_Tdict['VSC3600']
            elif p < 200 and 50 <= q <= 3500 and 50 <= h <= 2000:
                F_T = F_Tdict['VSC1800']
            elif p < 150 and 100 <= q <= 1500 and 100 <= h <= 450:
                F_T = F_Tdict['HSC3600']
            elif p < 250 and 250 <= q <= 5000 and 50 <= h <= 500:
                F_T = F_Tdict['HSC1800']
            elif p < 250 and 50 <= q <= 1100 and 300 <= h <= 1100:
                F_T = F_Tdict['2HSC3600']
            elif p < 1450 and 100 <= q <= 1500 and 650 <= h <= 3200:
                F_T = F_Tdict['2+HSC3600']
            else:
                raise NotImplementedError(f'no centrifugal pump available at current power ({p:.3g} hp), flow rate ({q:.3g} gpm), and head ({h:.3g} ft)')
            S = q*h**0.5 # Size factor
            S_new = S if S > 400 else 400
            lnS = ln(S_new)
            Cb = exp(12.1656-1.1448*lnS+0.0862*lnS**2)
            Cb *= S/S_new
            Cost['Pump'] = F_M*F_T*Cb*I
        elif Type == 'Gear':
            q_new = q if q > 50 else 50
            lnq = ln(q_new)
            Cb = exp(8.2816 - 0.2918*lnq + 0.0743*lnq**2)
            Cb *= q/q_new
            Cost['Pump'] = F_M*Cb*I
        elif Type == 'MeteringPlunger':
            Cb = exp(7.9361 + 0.26986*lnp + 0.06718*lnp**2)
            Cost['Pump'] = F_M*Cb*I
        
        # Cost electric motor
        lnp2 = lnp**2
        lnp3 = lnp2*lnp
        lnp4 = lnp3*lnp
        Cost['Motor'] = exp(5.9332 + 0.16829*lnp - 0.110056*lnp2 + 0.071413*lnp3 - 0.0063788*lnp4)*I
    
    @staticmethod
    def _nearest_PumpPower(p:'hp'):
        for power in nema_sizes_hp:
            if power >= p: return power
        return power
        
    @staticmethod
    def _calc_Efficiency(q:'gpm', p:'hp'):
        mup = _calc_BreakEfficiency(q)
        mum = _calc_MotorEfficiency(p/mup)
        return mup*mum
        
        